/*
** Module   : AnnotatedAst.java
** Abstract : an AST which supports annotations and other features needed for
**            post-processing
**
** Copyright (c) 2005-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 ECF 20050218   @19889 Created initial version.
** 002 ECF 20050221   @19895 Added annotationKeys() method. Exposes keys
**                           for all annotations (if any) in this AST.
** 003 ECF 20050221   @19903 Modified constructor for use by AstPersister.
**                           Token is now saved off for later use. Also
**                           made minor fix to getPath method and renamed
**                           fixupParents to fixups.
** 004 ECF 20050223   @19926 Added a copy constructor. Creates a deep copy
**                           of the original, including children. Added
**                           supporting implementation of initialize(AST),
**                           which overrides the superclass'
**                           implementation.
** 005 GES 20050223   @19929 Modified getFilename() processing to use the
**                           AstRegistry and added initialization of IDs
**                           to the fixups() method such that new ASTs
**                           will get their IDs assigned at the same
**                           time as parent nodes are fixed up.
** 006 ECF 20050224   @19949 Made all annotation keys lowercase. This
**                           ensures case-insensitive lookups. This
**                           implementation detail is exposed to the user
**                           however, when iterating annotation keys.
** 007 ECF 20050228   @20097 Added remove() method. Removes this node (and
**                           all descendant nodes) from containing tree.
** 008 GES 20050228   @20111 Added some helper methods to determine how
**                           many descendants (immediate and all) an AST
**                           has. Also added a token type symbol lookup.
** 009 ECF 20050303   @20161 Added two flavors of an iterator method. One
**                           walks the entire tree from the current node;
**                           the other walks only to a predetermined
**                           depth. The underlying implementation was
**                           moved from pattern.AstWalker.
** 010 ECF 20050303   @20168 Fixed defect in iterator method which would
**                           walk to the root's sibling if the root has no
**                           children.
** 011 ECF 20050316   @20417 Added notification mechanism inside AST walk
**                           iterator. Uses AstWalkListener to notify a
**                           listener of descents and ascents of nesting
**                           level traversals during a tree walk.
** 012 GES 20050318   @20432 Added dumpTree() (parser style output) and
**                           getAncestor() methods.
** 013 GES 20050323   @20476 Converted copy constructor into a split
**                           duplicate() method where the original object
**                           handles the construction of the new copy
**                           instance but the tough work of deep copying
**                           and recursion is done in the copy's worker
**                           routine.
** 014 GES 20050323   @20477 Converted to an abstract class and moved all
**                           token type/token name processing into the
**                           subclasses.
** 015 GES 20050325   @20507 Added a setParent() method to allow external
**                           fixups.
** 016 ECF 20050328   @20513 Added implementation of isRoot method.
** 017 GES 20050401   @20607 Added isAnnotation() method.
** 018 GES 20050406   @20653 Added support for array annotations.
** 019 ECF 20050408   @20658 Added graftCopyTo method. Grafts a copy of
**                           the current AST as a child of another AST,
**                           performing all necessary fixups. As part of
**                           the graft, unwanted nodes can be pruned from
**                           the copy.
** 020 GES 20050411   @20669 Added getImmediateChild() method to search
**                           for a child of a specific token type.
** 021 ECF 20050412   @20682 Corrected iteration in graftCopyTo method.
**                           Iteration of copy's children was mistakenly
**                           including the parent node itself.
** 022 GES 20050414   @20734 Important fix to the put worker for array
**                           annotations (it always was failing).
** 023 ECF 20050428   @20920 Added support for lateral (i.e. next-child)
**                           transition events during iteration. A walk
**                           listener is now notified when the walk
**                           transitions from a node to its next, right
**                           sibling, in order to mimic recursion.
** 024 GES 20050428   @20926 Added support for getting the number of
**                           immediate children that have a specific
**                           token type.
** 025 GES 20050501   @20955 Added a variant of getAncestor() which
**                           handles a search up the tree for an ancestor
**                           matching a specified token type.
** 026 ECF 20050525   @21285 Fixed defect with getPath method.
** 027 ECF 20050603   @21401 Added a variant of getImmediateChild() which
**                           uses a collection of types for its match. The
**                           original implementation now delegates to this
**                           one.
** 028 GES 20050603   @21421 Added getChildAt() method for index based
**                           child lookups.
** 029 GES 20050613   @21473 Added next child index support to the 
**                           iterator.
** 030 GES 20050620   @21525 Added graft() method which provides direct
**                           access to the part of the graftCopyTo()
**                           which is useful for adding a subtree without
**                           creating a copy first.
** 031 GES 20050627   @21552 Added graftAt() method which provides a way
**                           of inserting a child node at a specific
**                           child index.
** 032 GES 20050728   @21895 Added hidden mode support which changes
**                           iterator processing when requested.
** 033 ECF 20050817   @22158 Added a variant of putAnnotation() which adds
**                           a list of annotations at once.
** 034 NVS 20050825   @22249 Added an annotationCount() method that counts 
**                           annotations in an array that equal an object.
** 035 GES 20050921   @22780 Added nearestEnclosing() method to search
**                           for a common ancestor node given 2 other
**                           nodes in the same tree.  Added
**                           getIndexPos() and getPrevSibling() to
**                           provide easy access to siblings (these
**                           compliment getChildAt().
** 036 GES 20050928   @22870 Added move() to ease sorting and moving
**                           nodes in a tree or between trees.
** 037 GES 20051014   @23048 Added clearIds().
** 038 GES 20051027   @23130 Added getLastImmediateChild().
** 039 GES 20051113   @23321 Added ancestor/descendant methods.
** 040 ECF 20051115   @23334 Fixed defect in ancestor(int, int[]) variant.
**                           Current node is now ignored by test.
** 041 SIY 20051124   @23499 Added compareWith() method.
** 042 ECF 20060110   @23867 Added new variants of compareWith() and
**                           getAncestor(). New compareWith makes case
**                           sensitivity of text comparison optional. New
**                           getAncestor considers annotation values as
**                           well as type.
** 043 GES 20060129   @24136 Added new variant of duplicate and new
**                           convenience methods duplicateFresh.
** 044 GES 20060308   @24934 The parent must be set null during remove.
** 045 ECF 20060315   @25081 Fixed iterator. AstWalkListener was not being
**                           notified of the final ascent event in certain
**                           cases.
** 046 GES 20060405   @25382 Added clearAnnotations().
** 047 ECF 20060615   @27353 Added isAncestorOf(). Tests whether this node
**                           is another's ancestor.
** 048 ECF 20060818   @28688 Added toStringVerbose(). Primarily for
**                           debugging purposes.
** 049 GES 20060912   @29485 The put of a list-based annotations now
**                           replaces at the index specified rather than
**                           inserting (and pushing the other content
**                           "up" the list).
** 050 GES 20070531   @33949 Moved useful helper methods here from
**                           common AST support so that the code could
**                           be used in a wider variety of situations.
** 051 GES 20070629   @34333 Added setters for the line and column.
** 052 GES 20070629   @34343 Additional signature for graphCopyTo().
** 053 GES 20070720   @34647 Added getDescriptiveTokenText().  
** 054 ECF 20080730   @39273 Reduced memory footprint. Use smaller default
**                           sizes for annotations collections. Remove
**                           annotations collections when they become
**                           empty.
** 055 SIY 20090119   @41179 Added commonParent() method.
** 056 SIY 20090126   @41216 Merged methods commonParent() and  
**                           nearestEnclosing().
** 057 GES 20090421   @41826 Matched utility class package change.
** 058 GES 20090427   @42022 Import change.
** 059 GES 20090429   @42055 Match package and class name changes.
** 060 GES 20090515   @42225 Moved to AstManager from AstRegistry/AstPersister.
** 061 GES 20090518   @42367 Package change.
** 062 GES 20100927          Added support for tree-local data and shadow nodes.
** 063 GES 20101001          Shifted ShadowNode to private members. Fixed
**                           broken algorithm to properly link shadow nodes
**                           that are split between syntactical statements. The
**                           old approach duplicated AST nodes instead of
**                           properly linking them.
** 064 ECF 20130220          Performance improvement: refactored relativePath to retrieve only
**                           as much source path as is necessary for the comparison;
**                           re-implemented getPath (an expensive method) accordingly.
** 065 ECF 20130309          Further performance tweaks. Reworked getPath to require less
**                           expensive string operations; only lowercase annotation keys when
**                           necessary.
** 066 ECF 20140403          Implemented generics.
** 067 ECF 20140920          Reduced memory footprint by interning annotation keys always, and all
**                           stored strings if the backing storage medium is memory. Even though
**                           the values of strings vary considerably, there generally is enough
**                           duplication of common values to make this a worthwhile optimization.
**                           The values of annotation keys have less variety, which is why we
**                           intern them always. The trade-off of this memory optimization is some
**                           additional overhead for the intern operation, which has a native
**                           implementation.
** 068 OM  20150122          Implemented putAnnotation(Object) used by runtime evaluation.
** 069 OM  20150210          Added support for settable annotation output.
** 070 ECF 20150413          Added getAnnotationList method.
** 071 ECF 20150507          Minor javadoc fix.
*/

package com.goldencode.ast;

import java.io.*;
import java.util.*;
import java.util.List;
import antlr.*;
import antlr.collections.*;
import com.goldencode.util.*;

/**
 * A modified abstract syntax tree node which provides additional services.
 * Each instance of this class provides:
 * <ul>
 * <li>A persistent, unique identifier. This is implemented as a 64-bit
 *     value where the high-order double word represents a unique source
 *     file with which this AST is associated, and the low order double word
 *     represents a unique identifier within the AST hierarchy associated
 *     with that file.
 * <li>A link to a parent node. This permits leaf-to-root path traversal,
 *     which is necessary to support the pattern recognition engine. This
 *     engine uses this information both to walk the tree of ASTs during
 *     ruleset processing, and to provide certain user function and action
 *     features.
 * <li>An API to manage annotations used in post-processing. Annotations are
 *     used to attach arbitrary string, numeric, and boolean data to an AST
 *     for various reasons. These annotations are exposed by name as user
 *     expression variables during ruleset processing, and are available to
 *     user functions and actions.
 * <li>Access to source file name, line number and column, and nesting level
 *     information. This information is exposed via user expression variables
 *     during ruleset processing. It is made available to user functions and
 *     actions.
 * <li>Iterator processing to provide a powerful walking mechanism. This
 *     walk can be limited to a specific depth (or it can be unlimited). In
 *     addition, certain "non-visiting" events can be hooked during the walk,
 *     including descents to lower nesting levels, moves between children 
 *     (next-child events) and ascents back up to higher nesting levels. 
 *     Finally, the {@link #setHidden} flag can optionally be honored during
 *     the walk, to bypass child nodes marked as <code>hidden</code>. See
 *     {@link #iterator(int,AstWalkListener,boolean)}.
 * </ul>
 */
public abstract class AnnotatedAst
extends CommonAST
implements Aast 
{
   private static final boolean poolStrings;
   
   static
   {
      poolStrings = AstManager.get().usesMemoryStorage();
   }
   
   /** Unique, persistent identifier */
   private Long id = null;
   
   /** Source file line number associated with this AST */
   private int line = 0;
   
   /** Source file column number associated with this AST */
   private int column = 0;
   
   /** Parent node to this node */
   private Aast parent = null;
   
   /** Hidden mode which can be used to modify iterator based walking. */
   private boolean hidden = false;
   
   /** Map of annotation objects, indexed by name */
   private Map<String, Object> annotations = null;
   
   /** Tree-specific data (data that is shared by all nodes of a tree. */
   private TreeLocal local = null;
   
   /**
    * Default constructor.
    */
   public AnnotatedAst()
   {
      super();
   }
   
   /**
    * Constructor which accepts and wrappers token.
    *
    * @param   token
    *          Token upon which this node is based.
    */
   public AnnotatedAst(Token token)
   {
      super(token);
      
      extractTokenData(token);
   }
   
   /**
    * Translates a text representation of a token type into the
    * actual integer value.
    *
    * @param    tokenName
    *           The text representation of the token type.
    *
    * @return   A valid integer token type value or -1 if no match was found.
    */
   public abstract int lookupTokenType(String tokenName);
   
   /**
    * Translates an integer value of a token type into the associated 
    * human readable text representation.
    *
    * @param    type
    *           The integer token type.
    *
    * @return   A valid symbolic token type name or <code>null</code> if
    *           no match was found.
    */
   public abstract String lookupTokenName(int type);
   
   /**
    * Set the source file line number for this node.
    *
    * @param    line
    *           New line number to associate with this node.
    */
   public void setLine(int line)
   {
      this.line = line;
   }
   
   /**
    * Set the source file column number for this node.
    *
    * @param    column
    *           New column number to associate with this node.
    */
   public void setColumn(int column)
   {
      this.column = column;
   }
   
   /**
    * Grafts a copy of this AST node onto the specified parent AST. If a list
    * of valid token types is provided (i.e. <code>validTokens</code>), all
    * child nodes of the copy which <em>do not</em> match a valid token type
    * are removed from the copy node, before it is grafted onto the parent.
    * All necessary parent and ID fixups are performed on the copy to
    * integrate it with its new parent tree.
    * <p>
    * The line and column numbers will be preserved in the duplicated tree.
    *
    * @param   parent
    *          AST node which will become the direct parent of the grafted
    *          copy of this node.
    * @param   validTokens
    *          Optional array of valid token types. If not <code>null</code>,
    *          only those children of the copy node which match a valid token
    *          type are allowed to remain. All other child nodes are removed
    *          from the copy.
    *
    * @return  The new AST node grafted into the tree.
    *
    * @throws  AstException
    *          if any error occurs registering the grafted node with the AST
    *          registry.
    */
   public Aast graftCopyTo(Aast parent, int[] validTokens)
   throws AstException
   {
      return graftCopyTo(parent, validTokens, true);
   }
   
   /**
    * Grafts a copy of this AST node onto the specified parent AST. If a list
    * of valid token types is provided (i.e. <code>validTokens</code>), all
    * child nodes of the copy which <em>do not</em> match a valid token type
    * are removed from the copy node, before it is grafted onto the parent.
    * All necessary parent and ID fixups are performed on the copy to
    * integrate it with its new parent tree.
    *
    * @param   parent
    *          AST node which will become the direct parent of the grafted
    *          copy of this node.
    * @param   validTokens
    *          Optional array of valid token types. If not <code>null</code>,
    *          only those children of the copy node which match a valid token
    *          type are allowed to remain. All other child nodes are removed
    *          from the copy.
    * @param   preserve
    *          <code>true</code> to preserve line and column information in
    *          the duplicated tree.
    *
    * @return  The new AST node grafted into the tree.
    *
    * @throws  AstException
    *          if any error occurs registering the grafted node with the AST
    *          registry.
    */
   public Aast graftCopyTo(Aast parent, int[] validTokens, boolean preserve)
   throws AstException
   {
      AnnotatedAst copy = (AnnotatedAst) duplicate(null);
      
      // The duplicate method has already assigned parents correctly.
      // Iterate across one level of descendants and prune unwanted,
      // immediate children.
      if (validTokens != null)
      {
         // Copy array of valid types and sort in ascending order.
         int len = validTokens.length;
         int[] toks = new int[len];
         System.arraycopy(validTokens, 0, toks, 0, len);
         Arrays.sort(toks);
         
         // Collect nodes with invalid types. We can't remove them while
         // iterating, as this would disrupt the iteration.
         List<Aast> orphans = new ArrayList<>();
         Aast next = (Aast) copy.getFirstChild();
         while (next != null)
         {
            if (Arrays.binarySearch(toks, next.getType()) < 0)
            {
               orphans.add(next);
            }
            
            next = (Aast) next.getNextSibling();
         }
         
         // Remove each orphaned node.
         for (Aast orphan : orphans)
         {
            next = orphan;
            next.remove();
         }
      }
      
      // Reset all IDs to null in the copy to allow fixups to assign new IDs.
      Iterator<Aast> iter = copy.iterator();
      while (iter.hasNext())
      {
         AnnotatedAst next = (AnnotatedAst) iter.next();
         next.id = null;
         if (!preserve)
         {
            next.line   = 0;
            next.column = 0;
         }
      }
      
      // Graft the groomed copy onto the parent node and perform parent and ID fixups.
      parent.graft(copy);
      
      return copy;
   }
   
   /**
    * Grafts a new child subtree onto this AST. All necessary parent and ID 
    * fixups are performed to integrate the child tree with its new parent 
    * tree.
    *
    * @param   child
    *          AST subtree to add as a child of this node.
    *
    * @throws  AstException
    *          if any error occurs registering the grafted node with the AST
    *          registry.
    */
   public void graft(Aast child)
   throws AstException
   {
      // add the child subtree
      addChild(child);
      
      // drive fixups
      driveFixups(child);
   }
   
   /**
    * Grafts a new child subtree onto this AST at a specific 0-based child
    * index. All necessary parent and ID fixups are performed to integrate
    * the child tree with its new parent tree.
    *
    * @param   child
    *          AST subtree to add as a child of this node.
    * @param   index
    *          0 to insert as the first child, -1 to insert as the last
    *          child or a 0-based index to specify an exact child index
    *          at which to insert the new child.  All current children at
    *          that index or at a greater index, are moved right. If the
    *          index is greater than the current number of children, the
    *          new child will be added at the end.
    *
    * @throws  AstException
    *          if any error occurs registering the grafted node with the AST
    *          registry.
    */
   public void graftAt(Aast child, int index)
   throws AstException
   {
      if (index == 0)
      {
         // insert at the beginning
         child.setNextSibling(getFirstChild());
         setFirstChild(child);
      }
      else
      {
         Aast b4 = getChildAt(index - 1);
         
         // is this a valid index?
         if (b4 == null)
         {
            // add at end
            graft(child);
            
            // fixups are already done
            return;
         }
         
         child.setNextSibling(b4.getNextSibling());
         b4.setNextSibling(child);
      }
      
      // trigger fixups
      driveFixups(child);
   }
   
   /**
    * Creates a deep copy of the original, including all descendant ASTs. The
    * copy contains no cross-linked object references back to the original
    * AST. This method essentially results in recursive calls when
    * making copies of the AST's children.
    *
    * @return   The duplicated AST or <code>null</code> if there is any
    *           exception during the copy.
    */
   @Override
   public Aast duplicate()
   {
      return duplicate(null);
   }
   
   /**
    * Creates a deep copy of the original, including all descendant ASTs. The
    * copy contains no cross-linked object references back to the original
    * AST. This method essentially results in recursive calls when
    * making copies of the AST's children. If an ID map is provided, map this
    * new node to its ID.
    *
    * @param    idMap
    *           Map of IDs to ASTs. May be <code>null</code>.
    *
    * @return   The duplicated AST or <code>null</code> if there is any
    *           exception during the copy.
    */
   @Override
   public Aast duplicate(Map<Long, Aast> idMap)
   {
      AnnotatedAst dup = null;
      
      try
      {
         dup = this.getClass().newInstance();
      }
      
      catch (Exception exc)
      {
         // nothing to do but return
         return null;
      }
      
      // delegate the hard work to the copy
      dup.duplicateImpl(this, idMap);
      
      return dup;
   }
   
   /**
    * Creates a deep copy of the original, including all descendant ASTs. The
    * copy contains no cross-linked object references back to the original
    * AST. This method essentially results in recursive calls when
    * making copies of the AST's children.
    * <p>
    * All unique IDs of this new sub-tree are cleared before this method
    * returns (see {@link #clearIds}).
    *
    * @return   The duplicated AST or <code>null</code> if there is any
    *           exception during the copy.
    */
   @Override
   public Aast duplicateFresh()
   {
      Aast newTree = duplicate();
      newTree.clearIds();
      
      return newTree;
   }
   
   /**
    * Creates a deep copy of the original, including all descendant ASTs. The
    * copy contains no cross-linked object references back to the original
    * AST. This method essentially results in recursive calls when
    * making copies of the AST's children. If an ID map is provided, map this
    * new node to its ID.
    * <p>
    * All unique IDs of this new sub-tree are cleared before this method
    * returns (see {@link #clearIds}).
    *
    * @param    idMap
    *           Map of IDs to ASTs. May be <code>null</code>.
    *
    * @return   The duplicated AST or <code>null</code> if there is any
    *           exception during the copy.
    */
   @Override
   public Aast duplicateFresh(Map<Long, Aast> idMap)
   {
      Aast newTree = duplicate(idMap);
      newTree.clearIds();
      
      return newTree;
   }
   
   /**
    * Clears (sets to <code>null</code>) the unique ID of this node and all
    * descendant nodes using recursion.  Warning: this is very dangerous so
    * you have better know what you are doing!
    */
   @Override
   public void clearIds()
   {
      id = null;
      
      Aast child = (Aast) getFirstChild();
      
      while (child != null)
      {
         child.clearIds();
         child = (Aast) child.getNextSibling();
      }
   }
         
   /**
    * Clear all IDs of an AST and register it under a new filename. The
    * AST is walked to fix-up all parent-child relationships as part of
    * the re-initialization.
    * <p>
    * This would be done in order to safely save a copy AST under a
    * different filename than the source AST. It is necessary in order to
    * avoid having two ASTs registered to the same filename and ID in the
    * AST registry.
    * <p>
    * No shadow nodes will be created using this version of the method. All
    * hidden tokens will be dropped.
    *
    * @param   filename
    *          New filename under which the AST is registered.
    *
    * @throws  AstException
    *          if any error occurs registering AST or applying fixups.
    */
   @Override
   public void brainwash(String filename)
   throws AstException
   {
      brainwash(filename, false);
   }
   
   /**
    * Clear all IDs of an AST and register it under a new filename. The
    * AST is walked to fix-up all parent-child relationships as part of
    * the re-initialization.
    * <p>
    * This would be done in order to safely save a copy AST under a
    * different filename than the source AST. It is necessary in order to
    * avoid having two ASTs registered to the same filename and ID in the
    * AST registry.
    *
    * @param   filename
    *          New filename under which the AST is registered.
    * @param   shadow
    *          <code>true</code> to create shadow nodes from hidden tokens.
    *          <code>false</code> to drop hidden tokens without creating any
    *          shadow nodes.
    *
    * @throws  AstException
    *          if any error occurs registering AST or applying fixups.
    */
   public void brainwash(String filename, boolean shadow)
   throws AstException
   {
      this.clearIds();
      
      AstManager mgr = AstManager.get();
      Long       id  = mgr.addTree(filename);
      
      mgr.resetNextNodeId(id);
      this.fixups(null, id, shadow);
   }
   
   /**
    * Retrieve this AST's unique, persistent identifier. An ID should be
    * assigned after AST creation but before external post-processing. Thus,
    * a <code>null</code> return from this method generally indicates an
    * invalid AST.
    *
    * @return  A persistent value that uniquely identifies this AST node, or
    *          <code>null</code> if no ID has been assigned.
    */
   public Long getId()
   {
      return id;
   }
   
   /**
    * Indicate whether this AST node represents the root of its hierarchy.
    * This implementation defines "root-hood" as having a <code>null</code>
    * parent. Note that this will create false positives for an AST which
    * has not yet had its initial "fix-up" performed.
    *
    * @return  <code>true</code> if this node is the tree root, else
    *          <code>false</code>.
    */
   public boolean isRoot()
   {
      return (parent == null);
   }
   
   /**
    * Indicate whether this AST node represents a leaf node in the tree (a
    * node with no children).
    *
    * @return   <code>true</code> if this node has no children, else
    *           <code>false</code>.
    */
   public boolean isLeaf()
   {
      return (getFirstChild() == null);
   }
   
   /**
    * Set this AST's unique, persistent identifier. Once this value is set,
    * it is set "for life" and may not be changed.  An ID should be assigned
    * after AST creation but before external post-processing begins.
    * <p>
    * Normally, it would be better to set this value during construction.
    * However, we do not have control over the construction process when this
    * AST is instantiated by ANTLR during parsing, so it is done early in
    * post-processing instead.
    *
    * @param   id
    *          This AST's assigned ID.
    *
    * @return  <code>true</code> if the ID was set; <code>false</code> if it
    *          could not be set because it had previously been set.
    */
   public boolean setId(Long id)
   {
      if (this.id == null)
      {
         this.id = id;
         return true;
      }
      
      return false;
   }
   
   /**
    * Return a string representing the path of this AST from the root AST,
    * in terms of the token type of each element from the root to this AST.
    * The token types are separated by a forward slash and a trailing slash
    * ends the path. For example:
    * <pre>
    *    BLOCK/STATEMENT/KW_IF/
    * </pre>
    * represents the following path:
    * <pre>
    *    ROOT [type=BLOCK]
    *     |
    *    NODE A [type=STATEMENT]
    *     |
    *    NODE B [type=KW_IF]
    * </pre>
    * The string uniquely identifies the path from this node back to the root,
    * but is not necessarily unique when traversed from the root to this node.
    *
    * @return  A string which, when read from right to left, indicates the
    *          token type of each node from this AST to the root AST.
    */
   public String getPath()
   {
      return getPath(0);
   }
   
   /**
    * Get the level of nesting of this AST from the root AST, where
    * <code>0</code> indicates the root, and a positive number indicates
    * generations from the root.
    *
    * @return  Nesting level.
    */
   public int getDepth()
   {
      int depth = 0;
      for (Aast next = this; (next = next.getParent()) != null; depth++)
         ;
      return depth;
   }
   
   /**
    * Get the number of immediate child nodes of this AST.
    *
    * @return  Number of immediate child nodes.
    */
   public int getNumImmediateChildren()
   {
      int num = 0;
      
      Aast child = (Aast) getFirstChild();
      
      while (child != null)
      {
         num++;
         child = (Aast) child.getNextSibling();
      }
           
      return num;
   }
   
   /**
    * Get the number of immediate child nodes of this AST that match a
    * specific token type.
    *
    * @param    type
    *           The token type of the child nodes to search for.
    *
    * @return  Number of immediate child nodes which have the specified
    *          token type.
    */
   public int getNumImmediateChildren(int type)
   {
      int num = 0;
      
      Aast child = (Aast) getFirstChild();
      
      while (child != null)
      {
         if (child.getType() == type)
         {
            num++;
         }
         
         child = (Aast) child.getNextSibling();
      }
           
      return num;
   }   
   
   /**
    * Get the total number of all descendants of this AST.
    *
    * @return  Total number of descendants.
    */
   public int getNumChildren()
   {
      int num = 0;
      
      Aast child = (Aast) getFirstChild();
      
      while (child != null)
      {
         // count this child
         num++;
         
         // recursively ask this child how many it has
         num += child.getNumChildren();
         
         // get next child
         child = (Aast) child.getNextSibling();
      }
           
      return num;
   }
   
   /**
    * Initialize this node with the specified AST. Delegates to the superclass
    * and copies data from <code>ast</code>.
    *
    * @param   ast
    *          AST upon which this node is based. Expected to be an instance
    *          of {@link Aast}.
    */
   public void initialize(AST ast)
   {
      super.initialize(ast);
      
      if (ast instanceof Aast)
      {
         Aast aast = (Aast) ast;
         
         // Copy immutable data.
         this.line   = aast.getLine();
         this.column = aast.getColumn();
         this.id     = aast.getId();
         
         // Copy mutable data.
         this.hidden = aast.isHidden();
         this.local  = aast.getLocal();
         
         // Copy annotations map, if any.
         Iterator<String> iter = aast.annotationKeys();
         while (iter.hasNext())
         {
            String key = iter.next();
            putAnnotationImpl(key, aast.getAnnotation(key));
         }
      }
   }
   
   /**
    * Initialize this node with the specified token. Delegates to the
    * superclass and extracts data from <code>token</code> for later use in
    * line number and column queries.
    *
    * @param   token
    *          Token upon which this node is based.
    */
   public void initialize(Token token)
   {
      super.initialize(token);
      
      extractTokenData(token);
   }
   
   /**
    * Get this Aast's parent node, if any.
    *
    * @return  Parent Aast node, or <code>null</code> if this node represents
    *          the root of a hierarchy.
    */
   public Aast getParent()
   {
      return parent;
   }
   
   /**
    * Set this Aast's parent node.
    *
    * @param    parent
    *           Parent Aast node.
    */
   public void setParent(Aast parent)
   {
      this.parent = parent;
   }
   
   /**
    * Override the parent's implementation to pool text if so configured.
    * 
    * @param   text
    *          Token text for this node.
    */
   public void setText(String text)
   {
      super.setText((poolStrings && text != null) ? text.intern() : text);
   }
   
   /**
    * Get this AST's hidden status.
    *
    * @return   <code>true</code> if this node should be hidden during
    *           iterator based walks (in which the iterator is specified to 
    *           honor the hidden mode).
    */
   public boolean isHidden()
   {
      return hidden;
   }
   
   /**
    * Set this AST's hidden status.
    *
    * @param    hide
    *           <code>true</code> if this node should be hidden during
    *           iterator based walks (in which the iterator is specified to 
    *           honor the hidden mode).
    */
   public void setHidden(boolean hide)
   {
      hidden = hide;
   }
   
   /**
    * Returns the index position of this instance in the array of children 
    * of its parent.  For example, if this instance is the first child of
    * its parent, then this method will return 0.  If this child is the
    * second child of its parent, this method will return 1, as so on.
    *
    * @return  The 0-based index position of this node OR 0 if this node has
    *          no parent (i.e. it is the tree root).
    */
   public int getIndexPos()
   {
      if (parent == null)
      {
         return 0;
      }
      
      int sibs = 0;
      
      // start with the first child
      Aast child = (Aast) parent.getFirstChild();
      
      // check each sibling from the first until we get to ourselves
      while (child != this)
      {
         sibs++;
         child = (Aast) child.getNextSibling();
      }
      
      return sibs;
   }
      
   /**
    * Returns the sibling to the left of this node in the tree or 
    * <code>null</code> if this instance is the first child of its parent.
    * <p>
    * This method is much slower than its counterpart 
    * <code>getNextSibling</code> since each node does not maintain a
    * reference back to the previous node.  Thus, the parent node must
    * be involved and this extra processing slows the resulting
    * implementation.
    *
    * @return  The previous sibling or <code>null</code> if this node is the
    *          first child (or has no parent).
    */
   public Aast getPrevSibling()
   {
      // TODO: optimize this into a single pass search (rather than 2 
      //       traversals) OR change the base implementation to keep a
      //       reference back
      
      int pos = getIndexPos();
      
      if (pos == 0)
      {
         return null;
      }
      
      return parent.getChildAt(pos - 1);
   }
   
   /**
    * Get the ancestor node at the specified number of levels of ancestry,
    * if it exists.
    *
    * @param   level
    *          Number of generations up the tree to walk to obtain the 
    *          ancestor.  Set to <code>-1</code> to obtain the tree root, 
    *          <code>0</code> is the current node, <code>1</code> is the
    *          parent, <code>2</code> is the 'grandparent' and so forth.
    *
    * @return  The specified ancestor AST node if found or <code>null</code>
    *          if no such node exists (if <code>level</code> specified is 
    *          greater than depth of the tree).
    */
   public Aast getAncestor(int level)
   {
      // implement support for obtaining the root
      if (level == -1)
      {
         level = getDepth();
      }
      
      // start at the current node
      Aast result = this;
      
      // walk backwards through ancestors for level generations
      for (int i = 0; i < level && result != null; i++)
      {
         result = result.getParent();
      }
      
      return result;
   }
   
   /**
    * Get the first ancestor node of a specified type found in a walk up to
    * the maximum specified number of levels of ancestry, if such a node
    * exists.
    *
    * @param   level
    *          Maximum number of generations up the tree to search for an 
    *          ancestor.  Set to <code>-1</code> to search up to the tree 
    *          root, <code>0</code> for the current node, <code>1</code> for
    *          the parent, <code>2</code> for the 'grandparent' and so forth.
    * @param   type
    *          The token type of the ancestor to find.
    *
    * @return  The specified ancestor AST node if found or <code>null</code>
    *          if no such node exists (if <code>level</code> specified is 
    *          greater than depth of the tree).
    */
   public Aast getAncestor(int level, int type)
   {
      // implement support for obtaining the root
      if (level == -1)
      {
         level = getDepth();
      }
      
      // start at the current node
      if (getType() == type)
      {
         return this;
      }
      
      Aast result = this;
      
      // walk backwards through ancestors for level generations
      for (int i = 0; i < level && result != null; i++)
      {
         result = result.getParent();
         
         if (result != null && result.getType() == type)
         {
            return result;
         }
      }
      
      return null;
   }
   
   /**
    * Get the first ancestor node of the specified type which also has a
    * particular annotation with the specified value.  If a <code>null</code>
    * value is provided, then the existence alone of an annotation with the
    * specified key is considered a match.  The search walks up the tree from
    * this node to the maximum specified number of levels of ancestry.
    *
    * @param   level
    *          Number of generations up the tree to walk to obtain the 
    *          ancestor.  Set to <code>-1</code> to obtain the tree root, 
    *          <code>0</code> is the current node, <code>1</code> is the
    *          parent, <code>2</code> is the 'grandparent' and so forth.
    * @param   type
    *          The token type of the ancestor to find.
    * @param   key
    *          Annotation key to match.
    * @param   value
    *          Annotation value to match.  <code>null</code> to match any
    *          node which has an annotation with the specified key,
    *          regardless of its value.
    *
    * @return  The specified ancestor AST node if found or <code>null</code>
    *          if no such node exists (if <code>level</code> specified is 
    *          greater than depth of the tree).
    */
   public Aast getAncestor(int level, int type, String key, Object value)
   {
      // Implement support for obtaining the root.
      if (level == -1)
      {
         level = getDepth();
      }
      
      // Start at the current node.
      Aast result = this;
      
      // Walk backwards through ancestors for level generations.
      for (int i = 0; i <= level && result != null; i++)
      {
         // Match the type first.
         if (result.getType() == type)
         {
            // Match the annotation (the mere existence of the annotation is
            // considered a match if value is null).
            Object note = result.getAnnotation(key);
            if ((value == null && note != null) || (value != null && value.equals(note)))
            {
               return result;
            }
         }
         
         result = result.getParent();
      }
      
      return null;
   }
   
   /**
    * Searches the AST's ancestors for an exact token type match.  The
    * first match ends the search but in the case of no matches, all ancestor
    * nodes (up to the given maximum level) will have been searched.
    *
    * @param   level
    *          Maximum number of generations up the tree to search for an 
    *          ancestor.  Set to <code>-1</code> to search up to the tree 
    *          root, <code>0</code> for the current node, <code>1</code> for
    *          the parent, <code>2</code> for the 'grandparent' and so forth.
    * @param   type
    *          Token type to be matched.
    *
    * @return  <code>true</code> if <code>type</code> was found as a
    *          ancestor's token type within <code>level</code>
    *          generations, else <code>false</code>.
    */
   public boolean ancestor(int level, int type)
   {
      return ancestor(level, (new int[] { type }));
   }
      
   /**
    * Searches the AST's ancestors for a token type that matches a given
    * range of token types (inclusively).  The first match ends the 
    * search but in the case of no matches, all ancestor nodes (up to the 
    * given maximum level) will have been searched.
    *
    * @param   level
    *          Maximum number of generations up the tree to search for an 
    *          ancestor.  Set to <code>-1</code> to search up to the tree 
    *          root, <code>0</code> for the current node, <code>1</code> for
    *          the parent, <code>2</code> for the 'grandparent' and so forth.
    * @param   lower
    *          Lower boundary of the token type range.
    * @param   upper
    *          Upper boundary of the token type range.
    *
    * @return  <code>true</code> if a ancestor's token type was found to
    *          match a range of types within <code>level</code> generations,
    *          else <code>false</code>.
    */
   public boolean ancestor(int level, int lower, int upper)
   {
      Aast next = this;
      int  num  = 0;
      
      // walk upwards through ancestors and compare token types...
      // stop when the specified maximum level is reached (a negative level
      // means there is no limit) or when we run out of ancestors
      while (next != null && (level < 0 || num <= level))
      {
         int current = next.getType(); 
         
         if (current >= lower && current <= upper)
         {
            return true;
         }
                  
         // walk up one more level
         next = next.getParent();
         num++;
      }
      
      return false;
   }
   
   /**
    * Searches the AST's ancestors for a token type that matches any of a
    * given list of token types.  The first match ends the search but in the 
    * case of no matches, all ancestor nodes (up to the given maximum level)
    * will have been searched.  The current node (this) is omitted from the
    * search.
    *
    * @param   level
    *          Maximum number of generations up the tree to search for an 
    *          ancestor.  Set to <code>-1</code> to search up to the tree 
    *          root, <code>1</code> to search up to the parent, <code>2</code>
    *          <code>2</code> to search up to the 'grandparent', and so forth.
    * @param   types
    *          The list of token types for which to search.
    *
    * @return  <code>true</code> if a ancestor's token type was found to
    *          match any of the given list of types within <code>level</code>
    *          generations, else <code>false</code>.
    */
   public boolean ancestor(int level, int[] types)
   {
      if (level == 0)
      {
         level = 1;
      }
      
      Aast next = getParent();
      
      // walk upwards through ancestors and compare token types...
      // stop when the specified maximum level is reached (a negative level
      // means there is no limit) or when we run out of ancestors
      for (int num = 0; next != null && (level < 0 || num < level); num++)
      {
         int current = next.getType();
         
         for (int type : types)
         {
            if (current == type)
            {
               return true;
            }
         }
         
         // walk up one more level
         next = next.getParent();
      }
      
      return false;
   }
      
   /**
    * Searches the AST's descendants for an exact token type match.  The
    * first match ends the search but in the case of no matches, all child
    * nodes (up to the given maximum level) will have been searched.
    *
    * @param   level
    *          Number of generations to search before abandoning the test.
    *          Set to <code>0</code> to search exhaustively.
    * @param   type
    *          Token type to be matched.
    *
    * @return  <code>true</code> if <code>type</code> was found as a
    *          descendant's token type within <code>level</code>
    *          generations, else <code>false</code>.
    */
   public boolean descendant(int level, int type)
   {
      return descendant(level, (new int[] { type }));
   }
      
   /**
    * Searches the AST's descendants for a token type that matches a given
    * range of token types (inclusively).  The first match ends the 
    * search but in the case of no matches, all child nodes (up to the given
    * maximum level) will have been searched.
    *
    * @param   level
    *          Number of generations to search before abandoning the test.
    *          Set to <code>0</code> to search exhaustively.
    * @param   lower
    *          Lower boundary of the token type range.
    * @param   upper
    *          Upper boundary of the token type range.
    *
    * @return  <code>true</code> if a descendant's token type was found to
    *          match a range of types within <code>level</code> generations,
    *          else <code>false</code>.
    */
   public boolean descendant(int level, int lower, int upper)
   {
      Iterator<Aast> iter = iterator(level, null);
      
      // only interested in descendants, so skip the root
      iter.next();
      
      while (iter.hasNext())
      {
         Aast next = iter.next();
         
         int current = next.getType();
         
         if (current >= lower && current <= upper)
         {
            return true;
         }
      }
      
      return false;
   }
   
   /**
    * Searches the AST's descendants for a token type that matches any of a
    * given list of token types.  The first match ends the search but in the 
    * case of no matches, all child nodes (up to the given maximum level)
    * will have been searched.
    *
    * @param   level
    *          Number of generations to search before abandoning the test.
    *          Set to <code>0</code> to search exhaustively.
    * @param   types
    *          The list of token types for which to search.
    *
    * @return  <code>true</code> if a descendant's token type was found to
    *          match any of the given list of types within <code>level</code>
    *          generations, else <code>false</code>.
    */
   public boolean descendant(int level, int[] types)
   {
      Iterator<Aast> iter = iterator(level, null);
      
      // only interested in descendants, so skip the root
      iter.next();
      
      while (iter.hasNext())
      {
         Aast next = iter.next();
         
         int current = next.getType();
         
         for (int type : types)
         {
            if (current == type)
            {
               return true;
            }
         }
      }
      
      return false;
   }
   
   /**
    * Test whether any of this node's siblings has the given token type.
    *
    * @param    type
    *           Token type to match.
    *
    * @return   <code>true</code> if any of this node's siblings match, else
    *           <code>false</code>.
    */
   public boolean sibling(long type)
   {
      if (parent == null)
      {
         return false;
      }
      
      // iterate over all siblings and compare token type
      Iterator<Aast> iter = parent.iterator(1, null);
      
      // skip the parent, we are only interested in siblings
      iter.next();
      
      // walk siblings
      while (iter.hasNext())
      {
         Aast next = iter.next();
         
         // test the type of any node that is not us
         if (next != this && next.getType() == type)
         {
            return true;
         }
      }
      
      return false;
   }
   
   /**
    * Returns the number of siblings to this node's left (before this node
    * in a tree walk).  This is a synonym for {@link #getIndexPos}.
    *
    * @return   The number of left siblings or 0 if this node is the first
    *           child or has no parent (i.e. it is the tree root).
    */
   public int getNumLeftSiblings()
   {
      return getIndexPos();
   }
   
   /**
    * Returns the number of siblings to this node's right (after this node
    * in a tree walk).
    *
    * @return   The number of right siblings or 0 if this node is the last
    *           child or has no parent (i.e. it is the tree root).
    */
   public int getNumRightSiblings()
   {
      if (parent == null)
      {
         return 0;
      }
      
      int sibs = 0;
      
      Aast sibling = (Aast) getNextSibling();
      
      // check each sibling to the right until there are no more siblings
      while (sibling != null)
      {
         sibs++;
         sibling = (Aast) sibling.getNextSibling();
      }
      
      return sibs;
   }
   
   /**
    * Report whether this node is an ancestor of the given node.
    *
    * @param   level
    *          The number of generations to search before abandoning the test.
    *          Set to <code>0</code> to search exhaustively.
    * @param   other
    *          The node whose ancestry will be tested.
    *
    * @return  <code>true</code> if this node is an ancestor of
    *          <code>other</code>, else <code>false</code>.  Returns
    *          <code>false</code> if <code>this == other </code>.
    */
   public boolean isAncestorOf(int level, Aast other)
   {
      if (this == other)
      {
         return false;
      }
      
      Aast parent = other.getParent();
      for (int i = 0; parent != null && (level == 0 || i < level); i++)
      {
         if (this == parent)
         {
            return true;
         }
         
         parent = parent.getParent();
      }
      
      return false;
   }
   
   /**
    * Compares this node's complete path from the root node (down to and
    * including the current node) with the specified path.
    *
    * @param    path
    *           Root-to-current path, formatted as described in 
    *           {@link #getPath}.
    *
    * @return   <code>true</code> if specified path matches this node's path
    *           exactly, else <code>false</code>.
    */
   public boolean absolutePath(String path)
   {
      return getPath().equalsIgnoreCase(path);
   }
   
   /**
    * Determines if this node's path from the root <b>ends with</b> the
    * specified path, on a case-insensitive basis. 
    * <p>
    * This node's type IS included in this path, which is different from the
    * {@link #upPath} which compares paths upward starting at the parent
    * node.
    *
    * @param    path
    *           Partial path from this node up to an arbitrary number of
    *           parent nodes, formatted as described in {@link #getPath}.
    *
    * @return   <code>true</code> if specified path matches the end of this 
    *           node's path exactly, else <code>false</code>.
    */
   public boolean relativePath(String path)
   {
      int max = path.length();
      if (!path.endsWith("/"))
      {
         max++;
      }
      
      return searchUp(getPath(max), path);
   }
   
   /**
    * Determines if this node's path upward <b>matches</b> the specified
    * path, on a case-insensitive basis. The format of the 'segments' of the
    * up path starts with the parent's type and each subsequent segment must
    * be the type of the next parent up in the hierarchy.
    * <p>
    * The type of the current node is NOT included in the up path.  This
    * is the only difference between this method and {@link #relativePath}.
    *
    * @param    path
    *           Partial path from this node's parent upward to an arbitrary 
    *           number of ancestor nodes, formatted as described in
    *           {@link #getPath}.
    *
    * @return   <code>true</code> if specified path exactly matches the 
    *           upward path from this node's parent, else <code>false</code>.
    */
   public boolean upPath(String path)
   {
      if (parent == null)
      {
         return false;
      }
      
      return parent.relativePath(path);
   }
   
   /**
    * Determines if this node's path downward <b>matches</b> the specified
    * path, on a case-insensitive basis. More than 1 down path may match this
    * specification since sibling nodes can have the same token type. If any
    * such match exists then this function will return <code>true</code>.
    * <p>
    * The format of the 'segments' of the down path starts with the
    * immediate child's type and each subsequent segment must be the type of
    * the next child down in the hierarchy.  The type of the current node is
    * NOT included in the down path.
    *
    * @param    path
    *           Partial path from this node down to an arbitrary number
    *           of descendant nodes, formatted as described in {@link #getPath}.
    *
    * @return   <code>true</code> if specified path exactly matches the 
    *           downward path from this node, else <code>false</code>.
    */
   public boolean downPath(String path)
   {
      // we can't have a match if their are no children
      if (getNumImmediateChildren() == 0)
      {
         return false;
      }
      
      // make our list of matches
      String[] list = path.split("/");
      
      // search for each element of the down path in turn
      return searchDown(list, 0);
   }
   
   /**
    * Find and return the nearest ancestor of the current instance which also
    * encloses the node passed as a parameter.  By definition, if both nodes
    * exist in the same tree, there must be a common ancestor.  The only
    * exception to this rule is if one of the two nodes encloses the other
    * (e.g. it is the tree root or is itself the ancestor of the other node).
    * In this case, the enclosing node itself will be returned. If both
    * references are to the same node, then the parent node will be returned.
    *
    * @param    other
    *           Another reference node in the same tree.
    *
    * @return   The nearest enclosing ancestor or <code>null</code> if
    *           the two AST nodes are not in the same tree.  If the two
    *           references are identical (they are the same node), the
    *           parent node will be returned (even if this is 
    *           <code>null</code> due to the given node being the tree
    *           root). Finally, if one node encloses the other, the
    *           enclosing node will be returned.
    */
   public Aast nearestEnclosing(Aast other)
   {
      if (other == null)
      {
         return null;
      }
      
      // both nodes are the same!
      if (this == other)
      {
         // if we are the root, null will be returned here...
         return parent;
      }
      
      Aast[] chain1 = getParentChain(this);
      Aast[] chain2 = getParentChain(other);
      
      int sz = Math.min(chain1.length, chain2.length);
      
      for (int i = 0; i < sz; i++)
      {
         if (chain1[i] != chain2[i])
         {
            if (i == 0) // roots do not match
            {
               return null;
            }
            
            return chain1[i - 1];
         }
      }
      
      // one node is root for another, in this case shorter chain belongs to common root
      return (chain1.length < chain2.length) ? this : other;
   }
   
   /**
    * Search the immediate children for the one at the specified index. If no
    * such child exists, return <code>null</code>.
    *
    * @param    index
    *           The 0-based index at which to search.
    *
    * @return   The child at the specified index or <code>null</code> if no
    *           such index exists.
    */
   public Aast getChildAt(int index)
   {      
      Aast child = (Aast) getFirstChild();
      
      // a negative index is invalid
      if (index < 0)
      {
         child = null;
      }
      
      for (int i = 0; i < index && child != null; i++)
      {
         child = (Aast) child.getNextSibling();
      }
      
      return child;
   }
   
   /**
    * Search the immediate children for one that matches a specific token 
    * type.  An optional starting child AST can be provided which allows 
    * a search to be made for matches subsequent to the first, if the start
    * AST is <code>null</code>, then the search will start with the first
    * child.
    *
    * @param    type
    *           The token type of the child for which to search.
    * @param    start
    *           The child from which to start the search or <code>null</code>
    *           if the search is to start with the first child.
    *
    * @return   The first child (to the right of the starting point) that
    *           matches or <code>null</code> if no match was found.
    */
   public Aast getImmediateChild(int type, Aast start)
   {
      List<Integer> types = new ArrayList<>(1);
      types.add(type);
      
      return getImmediateChild(types, start);
   }
   
   /**
    * Search the immediate children for one that matches any token type in
    * the specified Collection of types.  The first match is returned, as
    * determined by the iteration order of the <code>types</code> collection.
    * An optional starting child AST can be provided which allows a search
    * to be made for matches subsequent to the first, if the start AST is
    * <code>null</code>, then the search will start with the first child.
    *
    * @param    types
    *           Collection of token types of the child for which to search.
    * @param    start
    *           The child from which to start the search or <code>null</code>
    *           if the search is to start with the first child.
    *
    * @return   The first child (to the right of the starting point) that
    *           matches or <code>null</code> if no match was found.
    */
   @Override
   public Aast getImmediateChild(Collection<Integer> types, Aast start)
   {
      Aast child;
      
      if (start == null)
      {
         // start the search with the first child
         child = (Aast) getFirstChild();
      }
      else
      {
         // start search with sibling to the right
         child = (Aast) start.getNextSibling();
      }
      
      // perform the search
      while (child != null)
      {
         for (Integer type : types)
         {
            if (child.getType() == type)
            {
               // we have a match!
               return child;
            }
         }
         
         // move rightwards through the immediate children
         child = (Aast) child.getNextSibling();
      }
      
      // no match
      return null;
   }
   
   /**
    * Search the immediate children for last child that matches a specific 
    * token type.
    *
    * @param    type
    *           The token type of the child for which to search.
    *
    * @return   The last child (rightmost) that matches or <code>null</code>
    *           if no match was found.
    */
   public Aast getLastImmediateChild(int type)
   {
      Aast last = null;
      Aast next = null;
      
      next = getImmediateChild(type, last);
      
      // if there are no matches, this drops out and returns null, otherwise
      // the loop will not exit until the last match is found
      while (next != null)
      {
         // remember the last child that matched
         last = next;
         
         // look for the next match
         next = getImmediateChild(type, last);
      }
      
      return last;
   }
   
   /**
    * Remove this AST (and the entire sub-tree rooted at this node) from its
    * parent's collection of children. Note that the implication of such
    * removal is that all descendants of this node remain parented by this
    * sub-tree but are removed from the original parent tree along with this
    * node.
    * <p>
    * This method has no effect if invoked on the root node of an AST.
    */
   public void remove()
   {
      if (parent == null)
      {
         return;
      }
      
      AST last = null;
      AST current = parent.getFirstChild();
      AST next = null;
      
      while (current != null)
      {
         next = current.getNextSibling();
         if (current == this)
         {
            if (last != null)
            {
               last.setNextSibling(next);
            }
            else
            {
               parent.setFirstChild(next);
            }
            break;
         }
         last = current;
         current = next;
      }
      
      // clear our links
      this.setNextSibling(null);
      parent = null;
   }
   
   /**
    * Move the node from one location to another.  The move can occur between
    * index positions under a single parent, between 2 different parent nodes
    * in the same tree or between 2 different parent nodes in different
    * trees.  The type of the <code>Aast</code> nodes must be the same
    * between the current node and the parent node, otherwise the results 
    * will be undefined.
    * <p>
    * If the given parent node is <code>null</code> and the current parent
    * node of this instance is <code>null</code>, this method returns
    * immediately since there is nothing to do.
    * 
    * @param    parent
    *           The parent node which is the target of the move operation or
    *           <code>null</code> if the move is only affecting the index
    *           position of this node (a move of the child to a different
    *           index in the same parent).
    * @param    index
    *           0 to insert as the first child, -1 to insert as the last
    *           child or a 0-based index to specify an exact child index
    *           at which to insert the new child.  All current nodes at
    *           that index or at a greater index, are moved right. If the
    *           index is greater than the current number of nodes, the node
    *           will be added at the end.
    *
    * @throws   AstException
    *           if any error occurs registering the grafted node with the AST
    *           registry.
    */
   public void move(Aast parent, int index)
   throws AstException
   {
      Aast target = parent == null ? this.parent : parent;
      
      if (target == null)
      {
         // nothing to do!
         return;
      }
      
      // detach our subtree from the current node
      this.remove();
      
      // add the node to the target
      target.graftAt(this, index);
   }
   
   /**
    * Get source code filename associated with the current node. This
    * information is retrieved from the {@link AstManager} by this AST's
    * ID.
    *
    * @return  Source code filename associated with this AST or 
    *          <code>null</code> if the ID is <code>null</code> or invalid.
    */
   public String getFilename()
   {
      if (id != null)
      {
         try
         {
            return AstManager.get().getTreeName(id);
         }
         catch (AstException ae)
         {
            // report the error and then continue, this way we will see when
            // there is some registry problem BUT we won't halt processing
            ae.printStackTrace();
         }
      }
      
      return null;  
   }
   
   /**
    * Get the source code line number which designates where in the source
    * code file the token wrappered by this node was found. This will always
    * return <code>0</code> for an artificially manufactured AST (i.e., one
    * not generated by a parser).
    *
    * @return  Source file line number where the token underlying this AST
    *          was found, or <code>0</code> for an artificial AST.
    *
    * @see     #getColumn
    */
   public int getLine()
   {
      return line;
   }
   
   /**
    * Get the source code column number which designates where in the source
    * code file the token wrappered by this node was found. This will always
    * return <code>0</code> for an artificially manufactured AST (i.e., one
    * not generated by a parser).
    *
    * @return  Column number within a specific source code line, indicating
    *          where the token underlying this AST was found, or
    *          <code>0</code> for an artificial AST.
    *
    * @see     #getLine
    */
   public int getColumn()
   {
      return column;
   }
   
   /**
    * Returns the symbolic representation of this AST's token type.
    *
    * @return  Symbolic token name for this AST's token type.
    */
   public abstract String getSymbolicTokenType();
   
   /**
    * Returns a descriptive string for the current node.  This string will
    * normally contain both the text and the symbolic representation of this
    * AST's token type.
    *
    * @return   A descriptive string for the token type of this node.
    */
   public String getDescriptiveTokenText()
   {
      return String.format("%s [%s]", getText(), getSymbolicTokenType());
   }
   
   /**
    * Return an iterator to walk this AST recursively in a depth-first
    * traversal. 'Depth-first' means that the iteration will traverse the
    * children of any given node before traversing that node's siblings.
    * <p>
    * This iterator will not support any "non-visiting" events (descent,
    * ascent and next-child).  In addition, all nodes will be returned,
    * regardless of hidden status.   EXCEPTION: shadow nodes are NOT included
    * in any tree walks and will NOT be returned by this iterator.
    *
    * @return  An iterator which will walk this node and all of its
    *          descendants.
    */
   @Override
   public Iterator<Aast> iterator()
   {
      return iterator(0, null);
   }
   
   /**
    * Return an iterator to walk this AST recursively in a depth-first
    * traversal. 'Depth-first' means that the iteration will traverse the
    * children of any given node before traversing that node's siblings.
    * <p>
    * The iterator will walk at most to a node depth of <code>maxLevels</code>
    * relative to this node. A <code>maxLevels</code> value of <code>0</code>
    * indicates that the full tree beneath this node should be walked, down
    * to every leaf, regardless of its depth.
    * <p>
    * If a <code>listener</code> is specified, it will be notified when the
    * tree walk transitions between nesting levels. The notifications occur
    * during processing of the iterator's <code>next</code> method. In each
    * case (whether the transition represents a descent to a lower tree level
    * or an ascent to a higher tree level), the parent of the parent/child
    * link being traversed will be sent with the notification. No more than
    * a single descent notification will ever be sent during a single
    * invocation of the iterator's <code>next</code> method. However,
    * multiple ascent notifications may be sent, since the tree walk may
    * proceed directly from a deeply nested node to the next sibling of a
    * distant ancestor node. In such a case, one call to the listener's
    * {@link AstWalkListener#ascent} method will be made for each upward
    * step from child to parent, until an ancestor with an unvisited sibling
    * is found, or the walk ends.
    * <p>
    * This iterator will return all nodes in the tree, regardless of hidden
    * status.  EXCEPTION: shadow nodes are NOT included in any tree walks and
    * will NOT be returned by this iterator.
    *
    * @param   maxLevels
    *          Maximum number of levels or generations to visit relative to
    *          the root defined by <code>ast</code>. If the tree contains
    *          more levels than <code>maxLevels</code>, the levels beyond a
    *          relative depth of <code>maxLevels</code> will be ignored. Set
    *          to <code>0</code> to disable this restriction. Set to
    *          <code>1</code> for immediate children only, etc.
    * @param   listener
    *          An optional listener which is notified of significant events
    *          during the tree walk, such as descents to lower nesting levels,
    *          moves between children (next-child events) and ascents back up
    *          to higher nesting levels. This may be <code>null</code> if
    *          these "non-visiting" events are not needed.
    *
    * @return  An iterator which will behave as described above.
    */
   @Override
   public Iterator<Aast> iterator(final int             maxLevels,
                                  final AstWalkListener listener)
   {
      return iterator(maxLevels, listener, false);
   }
   
   /**
    * Return an iterator to walk this AST recursively in a depth-first
    * traversal. "Depth-first" means that the iteration will traverse the
    * children of any given node before traversing that node's siblings.
    * <p>
    * The iterator will walk at most to a node depth of <code>maxLevels</code>
    * relative to this node. A <code>maxLevels</code> value of <code>0</code>
    * indicates that the full tree beneath this node should be walked, down
    * to every leaf, regardless of its depth.
    * <p>
    * If a <code>listener</code> is specified, it will be notified when the
    * tree walk transitions between nesting levels. The notifications occur
    * during processing of the iterator's <code>next</code> method. In each
    * case (whether the transition represents a descent to a lower tree level
    * or an ascent to a higher tree level), the parent of the parent/child
    * link being traversed will be sent with the notification. No more than
    * a single descent notification will ever be sent during a single
    * invocation of the iterator's <code>next</code> method. However,
    * multiple ascent notifications may be sent, since the tree walk may
    * proceed directly from a deeply nested node to the next sibling of a
    * distant ancestor node. In such a case, one call to the listener's
    * {@link AstWalkListener#ascent} method will be made for each upward
    * step from child to parent, until an ancestor with an unvisited sibling
    * is found, or the walk ends.
    * <p>
    * Optionally, the returned iterator can omit nodes in which the hidden
    * status is enabled (see {@link #isHidden} and {@link #setHidden}).
    * Please note that if the root node (the instance upon which this method
    * is called) is set as <code>hidden</code>, this setting will
    * <code>not</code> be honored.  The <code>hidden</code> mode is only
    * honored on child nodes that have this mode set.
    * <p>
    * EXCEPTION: shadow nodes are NOT included in any tree walks and will NOT
    * be returned by this iterator.
    *
    * @param   maxLevels
    *          Maximum number of levels or generations to visit relative to
    *          the root defined by <code>ast</code>. If the tree contains
    *          more levels than <code>maxLevels</code>, the levels beyond a
    *          relative depth of <code>maxLevels</code> will be ignored. Set
    *          to <code>0</code> to disable this restriction. Set to
    *          <code>1</code> for immediate children only, etc.
    * @param   listener
    *          An optional listener which is notified of significant events
    *          during the tree walk, such as descents to lower nesting levels,
    *          moves between children (next-child events) and ascents back up
    *          to higher nesting levels. This may be <code>null</code> if
    *          these "non-visiting" events are not needed.
    * @param   honorHidden
    *          If <code>true</code>, the <code>hidden</code> attribute of
    *          the node will be honored.  This means that the node will not
    *          appear in the iterator nor will any "non-visiting" events be
    *          triggered for such nodes.
    *
    * @return  An iterator which will behave as described above.
    */
   @Override
   public Iterator<Aast> iterator(final int             maxLevels,
                                  final AstWalkListener listener,
                                  final boolean         honorHidden)
   {
      return new Iterator<Aast>()
      {
         /** The next AST to visit */
         private Aast next = AnnotatedAst.this;
         
         /** Parent of the current AST */
         private Aast parent = getParent();
         
         /** Current node depth from the starting node */
         private int level = 0;
         
         /** Direction of last level jump: -1 for descent, +1 for ascent */
         private int jump = 0;
         
         /** Whether a lateral (next-child) notification is due */
         private boolean notifyLateral = false;
         
         /** List of parent nodes to send with jump notifications */
         private List<Aast> parents = new ArrayList<>();
         
         /** Track indices of the next child to be processed. */
         private java.util.Stack<Integer> stack = new java.util.Stack<>();
         
         /**
          * Indicate whether the iterator has any more elements to visit.
          * It is assumed that this method is invoked at least once before
          * each call to the <code>next()</code> method, as is common in
          * most uses, where this method is called as a loop test, and
          * <code>next()</code> is called inside the loop. <strong>If this
          * is not the case, the final notification to any listener of
          * ascent events back to the root of the tree from the last node
          * <em>will not occur</em>.</strong>
          *
          * @return  <code>true</code> if more elements exist, else
          *          <code>false</code>.
          */
         public boolean hasNext()
         {
            boolean hasNext = (next != null);
            
            // Notify listener of final ascents if at the end of the walk.
            if (!hasNext)
            {
               notifyListenerLevelChanged();
            }
            
            return hasNext;
         }
         
         /**
          * Return the next element in the iteration sequence. If a listener
          * is attached to the iteration, it is notified of vertical and
          * lateral transitions between node visits.
          *
          * @return  Next element.
          *
          * @throws  NoSuchElementException
          *          if no further elements exist.
          */
         public Aast next()
         {
            if (next == null)
            {
               throw new NoSuchElementException();
            }
            
            // Notify listener of any nesting level transitions which
            // occurred the last time we advanced to the next node.
            notifyListenerLevelChanged();
            
            // Notify listener if a lateral transition occurred the last
            // time we advanced to the next node. Send the next node's
            // parent with the event.
            if (listener != null && notifyLateral)
            {
               // Determine the next child index to be processed.
               int childIdx = stack.peek() + 1;
               int size     = stack.size();
               
               // Invoke the listener for the nextChild event. 
               listener.nextChild(next.getParent(), childIdx);
               
               // Store the updated next child index in our stack.
               stack.setElementAt(childIdx, (size - 1));
               
               notifyLateral = false;
            }
            
            Aast ast = next;
            
            // Advance to the next node.
            advanceNext();
            
            return ast;
         }
         
         /**
          * Unsupported remove method.
          *
          * @throws  UnsupportedOperationException
          *          always.
          */
         public void remove()
         {
            throw new UnsupportedOperationException();
         }
         
         /**
          * Notify listener of any nesting level transitions which occurred
          * the last time we advanced to the next node. Notification events
          * are queued up when we advance to the next node, but are not sent
          * until the next node is requested. If multiple level changes have
          * taken place (such as when transitioning from a child node to a
          * great uncle node, for example), the list of parents traversed
          * is stored in the <code>parents</code> list, such that each
          * notification can be sent to the listener with the correct parent
          * in scope.
          */
         private void notifyListenerLevelChanged()
         {
            // Abort if nothing to do.
            if (listener == null || jump == 0)
            {
               return;
            }
            
            // Walk through list of parents and send appropriate notification for each one.
            for (Aast ast : parents)
            {
               if (jump > 0)
               {
                  // Notify listener of an ascent, sending the parent of
                  // the parent-child link being traversed.
                  listener.ascent(ast);
                  
                  // Pop a scope off our stack of next child indices.
                  stack.pop();
               }
               else if (jump < 0)
               {
                  // Push a scope on to our stack of next child indices.
                  stack.push(0);
                  
                  // Notify listener of a descent, sending the parent of
                  // the parent-child link being traversed.
                  listener.descent(ast);
               }
            }
            
            // Reset state variables for the next node advancement.
            parents.clear();
            jump = 0;
         }
         
         /**
          * Set the next element to be visited in the depth-first traversal.
          * The next element will be set to <code>null</code> if the full
          * traversal has completed. This method is responsible for setting
          * state variables as nesting levels in the tree are traversed, so
          * that the <code>next()</code> method can notify a listener of
          * these transitions.
          * <p>
          * If instructed to honor the hidden mode, any movement between
          * levels or between siblings will first check to see if the
          * <code>hidden</code> flag is set.  In cases where this flag is
          * <code>true</code>, those nodes will be treated as if they do
          * not exist for purposes of the iterator (and all "non-visiting"
          * events that are generated).
          */
         private void advanceNext()
         {
            AnnotatedAst current = (AnnotatedAst) next;
            next = null;
            
            // Determine if we're allowed to drill down to the next level.
            if (maxLevels == 0 || level < maxLevels)
            {
               next = current.getFirstChild(honorHidden);
            }
            
            if (next != null)
            {
               // We moved down a level in the hierarchy. Reset the parent
               // accordingly and increment our level count.
               parent = current;
               level++;
               
               // Store parent to use when notifying listener of the descent
               // on the next call to next(). Set jump direction to indicate
               // descent.
               if (listener != null && parent != null)
               {
                  parents.add(parent);
                  jump = -1;
               }
            }
            else if (level > 0)
            {
               // We either were not allowed to drill down an additional
               // level in the hierarchy, or there was no level to which to
               // descend. So, try to visit the next sibling at this level.
               next = current.getNextSibling(honorHidden);
               
               // If we found a sibling to our right, trip a flag indicating
               // that a next-child notification must be sent the next time
               // the "next" method is called.
               if (next != null)
               {
                  notifyLateral = true;
               }
            }
            
            // If we found no more siblings at the current level, try
            // ascending back one level and visiting the parent's next
            // sibling. Keep ascending until we find one or we end up back
            // at the starting level. DO NOT advance to a sibling of the
            // starting node.
            while (next == null && parent != null && --level > 0)
            {
               // Store parent to use when notifying listener of the ascent(s)
               // on the next call to next(). Set jump direction to indicate
               // ascent.
               if (listener != null)
               {
                  parents.add(parent);
                  jump = 1;
               }
               
               next = ((AnnotatedAst) parent).getNextSibling(honorHidden);
               
               // If we found a sibling to our right, trip a flag indicating
               // that a next-child notification must be sent the next time
               // the "next" method is called.
               if (next != null)
               {
                  notifyLateral = true;
               }
               
               parent = parent.getParent();
            }
            
            // Handle final ascent to starting node.
            if (listener != null && next == null && parent != null)
            {
               parents.add(parent);
               jump = 1;
            }
         }
      };
   }
   
   /**
    * Determines if an annotation exists of any <code>String</code>,
    * <code>Double</code>, <code>Long</code>, <code>Boolean</code> or
    * <code>List</code> value, indexed by an arbitrary key.
    *
    * @param   key
    *          Key under which the annotation is stored.
    *
    * @return  <code>true</code> if an annotation exists mapped by that key.
    */
   public boolean isAnnotation(String key)
   {
      if (annotations == null)
      {
         return false;
      }
      
      return annotations.get(key.toLowerCase()) != null;
   }
   
   /**
    * Returns the size of an array annotation as indexed by an arbitrary key.
    *
    * @param   key
    *          Key under which the annotation is stored.
    *
    * @return  The number of elements in an array annotation that exists
    *          mapped by that key or 0 if no such named array annotation
    *          exists.  A scalar annotation using that same name will also
    *          result in a return of 0.
    */
   public int annotationSize(String key)
   {
      if (annotations == null)
      {
         return 0;
      }
      
      Object obj = annotations.get(key.toLowerCase());
      
      if (obj == null || !(obj instanceof List<?>))
      {
         return 0;
      }
      
      List<Object> array = (List<Object>) obj;
      
      return array.size();
   }
   
   /**
    * Returns the number of annotation objects of an array annotation that are
    * equal to the specified <code>Object</code>.
    *
    * @param   key
    *          Key under which the annotation is stored.
    *
    * @param   value
    *          <code>Object</code> to compare with.
    *
    * @return  The number of elements in an array annotation that equals
    *          the specified value or 0 if no such named array annotation
    *          exists.  A scalar annotation using that same name will also
    *          result in a return of 0.
    */
   public int annotationCount(String key, Object value)
   {
      Object obj = getAnnotation(key);
      
      if (obj == null || !(obj instanceof List<?>))
      {
         return 0;
      }
      
      List<Object> array = (List<Object>) obj;
      Iterator<Object> iter = array.iterator();
      int count = 0;
      
      while (iter.hasNext())
      {
         obj = iter.next();
         if (obj.equals(value))
         {
            count ++;
         }
      }
      
      return count;
   }
   
   /**
    * Retrieve an annotation. This may be any <code>String</code>,
    * <code>Double</code>, <code>Long</code>, or <code>Boolean</code> value,
    * indexed by an arbitrary key.  Note that it is possible to obtain the
    * entire array if the annotation is an array type.  This would be a
    * <code>List</code> of objects of the above mentioned types.
    *
    * @param   key
    *          Key under which the annotation is stored.
    *
    * @return  The annotation object stored under the specified key, or
    *          <code>null</code> if none exists.
    */
   public Object getAnnotation(String key)
   {
      if (annotations == null)
      {
         return null;
      }
      
      // try given key first, as lowercasing is expensive and most TRPL code which looks up an
      // annotation uses lowercase keys;  this is safe because we always lowercase the keys when
      // adding annotations
      Object obj = annotations.get(key);
      if (obj == null)
      {
         obj = annotations.get(key.toLowerCase());
      }
      
      return obj;
   }
   
   /**
    * Retrieve an element in an array annotation using the specified key as
    * a unique identifier and an index to specify the location in the array
    * from which to read the element.  The special index of -1 indicates the
    * last element in the array. The resulting data may be any
    * <code>String</code>, <code>Double</code>, <code>Long</code>, or
    * <code>Boolean</code> value.
    *
    * @param   key
    *          Key under which the annotation is stored.
    * @param   index
    *          The index of -1 indicates the last element of the array.
    *          Otherwise this is the explicit zero-based index at which to 
    *          find the annotation.
    *
    * @return  The annotation object stored under the specified key and at
    *          the specified index, or <code>null</code> if the annotation
    *          doesn't exist, if the annotation is not an array or if no
    *          element exists at the given index.
    */
   public Object getAnnotation(String key, int index)
   {
      List<Object> array = getAnnotationList(key);
      
      if (array == null)
      {
         return null;
      }
      
      if (index == -1)
      {
         // get the last element in the array
         index = array.size() - 1;
      }
      
      Object result = null;
      
      try
      {
         result = array.get(index);
      }
      catch (IndexOutOfBoundsException iob)
      {
         // eat the exception as we will simply return null
      }
      
      return result;
   }
   
   /**
    * Retrieve an unmodifiable list view of an array annotation.
    * 
    * @param   key
    *          Key under which the annotation is stored.
    * 
    * @return  Unmodifiable list view of the array annotation, or <code>null</code> if no such
    *          annotation exists.
    */
   public List<Object> getAnnotationList(String key)
   {
      if (annotations == null)
      {
         return null;
      }
      
      Object obj = getAnnotation(key);
      
      if (obj == null || !(obj instanceof List<?>))
      {
         return null;
      }
      
      List<Object> list = (List<Object>) obj;
      
      return Collections.unmodifiableList(list);
   }
   
   /**
    * Store a text annotation using the specified key as a unique identifier.
    * <p>
    * This is just a thin wrapper to ensure annotation type safety, since
    * annotations are stored in a <code>Map</code> as <code>Object</code>s.
    *
    * @param   key
    *          Key under which to store the annotation.
    * @param   annotation
    *          Arbitrary text.
    */
   public void putAnnotation(String key, String annotation)
   {
      putAnnotationImpl(key, annotation);
   }
   
   /**
    * Store a text element in an array annotation using the specified key as
    * a unique identifier and an index to specify the location in which to
    * insert.  If an element already exists at that index, it and all
    * following elements will be located at indices that are 1 greater than
    * before.  The special index of -1 indicates an addition to the end of
    * the array.
    *
    * @param   key
    *          Key under which to store the annotation.
    * @param   index
    *          The index of -1 indicates an addition to the end of the array.
    *          Otherwise this is the explicit zero-based index at which to 
    *          add the annotation.
    * @param   annotation
    *          Arbitrary text.
    */
   public void putAnnotation(String key, String annotation, int index)
   {
      putAnnotationImpl(key, annotation, index);
   }
   
   /**
    * Store a floating point numeric annotation using the specified key as a
    * unique identifier.
    * <p>
    * This is just a thin wrapper to ensure annotation type safety, since
    * annotations are stored in a <code>Map</code> as <code>Object</code>s.
    *
    * @param   key
    *          Key under which to store the annotation.
    * @param   annotation
    *          Arbitrary floating point numeric value.
    */
   public void putAnnotation(String key, Double annotation)
   {
      putAnnotationImpl(key, annotation);
   }
   
   /**
    * Store a floating point numeric element in an array annotation using the
    * specified key as a unique identifier and an index to specify the
    * location in which to insert.  If an element already exists at that
    * index, it and all following elements will be located at indices that
    * are 1 greater than before.  The special index of -1 indicates an
    * addition to the end of the array.
    *
    * @param   key
    *          Key under which to store the annotation.
    * @param   index
    *          The index of -1 indicates an addition to the end of the array.
    *          Otherwise this is the explicit zero-based index at which to 
    *          add the annotation.
    * @param   annotation
    *          Arbitrary floating point numeric value.
    */
   public void putAnnotation(String key, Double annotation, int index)
   {
      putAnnotationImpl(key, annotation, index);
   }
   
   /**
    * Store an integral numeric annotation using the specified key as a
    * unique identifier.
    * <p>
    * This is just a thin wrapper to ensure annotation type safety, since
    * annotations are stored in a <code>Map</code> as <code>Object</code>s.
    *
    * @param   key
    *          Key under which to store the annotation.
    * @param   annotation
    *          Arbitrary integral numeric value.
    */
   public void putAnnotation(String key, Long annotation)
   {
      putAnnotationImpl(key, annotation);
   }
   
   /**
    * Store an integral numeric element in an array annotation using the
    * specified key as a unique identifier and an index to specify the
    * location in which to insert.  If an element already exists at that
    * index, it and all following elements will be located at indices that
    * are 1 greater than before.  The special index of -1 indicates an
    * addition to the end of the array.
    *
    * @param   key
    *          Key under which to store the annotation.
    * @param   index
    *          The index of -1 indicates an addition to the end of the array.
    *          Otherwise this is the explicit zero-based index at which to 
    *          add the annotation.
    * @param   annotation
    *          Arbitrary integral numeric value.
    */
   public void putAnnotation(String key, Long annotation, int index)
   {
      putAnnotationImpl(key, annotation, index);
   }
   
   /**
    * Store a boolean annotation using the specified key as a unique
    * identifier.
    * <p>
    * This is just a thin wrapper to ensure annotation type safety, since
    * annotations are stored in a <code>Map</code> as <code>Object</code>s.
    *
    * @param   key
    *          Key under which to store the annotation.
    * @param   annotation
    *          Arbitrary boolean value.
    */
   public void putAnnotation(String key, Boolean annotation)
   {
      putAnnotationImpl(key, annotation);
   }
   
   /**
    * Store a boolean element in an array annotation using the
    * specified key as a unique identifier and an index to specify the
    * location in which to insert.  If an element already exists at that
    * index, it and all following elements will be located at indices that
    * are 1 greater than before.  The special index of -1 indicates an
    * addition to the end of the array.
    *
    * @param   key
    *          Key under which to store the annotation.
    * @param   index
    *          The index of -1 indicates an addition to the end of the array.
    *          Otherwise this is the explicit zero-based index at which to 
    *          add the annotation.
    * @param   annotation
    *          Arbitrary boolean value.
    */
   public void putAnnotation(String key, Boolean annotation, int index)
   {
      putAnnotationImpl(key, annotation, index);
   }
   
   /**
    * Store a list of annotations at once using the specified key as a unique
    * identifier.
    * The list should be made of objects of the same type, which, in turn,
    * should be one of the supported annotation types:
    * <ul>
    *    <li><code>String</code>
    *    <li><code>Double</code>
    *    <li><code>Long</code>
    *    <li><code>Boolean</code>
    * </ul>
    *
    * @param   key
    *          Key under which to store the annotation.
    * @param   annotation
    *          list of annotations of the same, supported type.
    */
   @Override
   public void putAnnotation(String key, List<Object> annotation)
   {
      Class<?> type = null;
      for (Object next : annotation)
      {
         Class<?> nextType = next.getClass();
         if (type == null)
         {
            type = nextType;
            
            if (type != String.class &&
                type != Long.class &&
                type != Double.class &&
                type != Boolean.class)
            {
               throw new IllegalArgumentException(
                  "Unsupported annotation type:  " + type.getName());
            }
         }
         
         if (type != nextType)
         {
            throw new IllegalArgumentException(
               "Inconsistent annotation type:  " + nextType.getName());
         }
         
         putAnnotationImpl(key, next, -1);
      }
   }
   
   /**
    * Store any object as annotation for this node.
    * It will work only with <code>InMemoryRegistryPlugin</code> since this is the only plugin
    * that does not truly serialize the nodes. The other plugins require the the annotation
    * objects to be serializable so only a small set of objects are supported.
    *
    * @param   key
    *          The name of the annotation.
    * @param   annotation
    *          The value for the annotation. Any type of object can be stored as long it is not
    *          required to be saved.
    */
   @Override
   public void putAnnotationObject(String key, Object annotation)
   {
      putAnnotationImpl(key, annotation);
   }
   
   /**
    * Remove a specific annotation from this AST.  Note that this annotation
    * may be an entire array (of type <code>List</code>) or it may be a
    * scalar value.
    *
    * @param   key
    *          Key under which the annotation is stored.
    *
    * @return  The annotation which was removed, or <code>null</code> if none was found.
    */
   public Object removeAnnotation(String key)
   {
      if (annotations == null)
      {
         return null;
      }
      
      Object note = annotations.remove(key.toLowerCase());
      
      if (annotations.isEmpty())
      {
         annotations = null;
      }
      
      return note;
   }
   
   /**
    * Remove a specific element from the array annotation in this AST.
    * The special index of -1 indicates the removal of the last element of
    * the array.
    *
    * @param   key
    *          Key under which the annotation is stored.
    * @param   index
    *          The index of -1 indicates an addition to the end of the array.
    *          Otherwise this is the explicit zero-based index at which to 
    *          add the annotation.
    *
    * @return  The element which was removed, or <code>null</code> if the
    *          annotation doesn't exist, if the annotation is not an array or
    *          if no element exists at the given index.
    */
   public Object removeAnnotation(String key, int index)
   {
      if (annotations == null)
      {
         return null;
      }
      
      String lcKey = key.toLowerCase();
      Object obj = annotations.get(lcKey);
      
      if (obj == null || !(obj instanceof List<?>))
      {
         return null;
      }
      
      List<Object> array = (List<Object>) obj;
      
      if (index == -1)
      {
         // get the last element in the array
         index = array.size() - 1;
      }
      
      Object result = null;
      
      try
      {
         result = array.remove(index); 
         
         if (array.isEmpty())
         {
            annotations.remove(lcKey);
            
            if (annotations.isEmpty())
            {
               annotations = null;
            }
         }
      }
      
      catch (IndexOutOfBoundsException iob)
      {
         // eat the exception as we will simply return null
      }
      
      return result; 
   }
   
   /**
    * Removes ALL annotations from this AST.  BE VERY CAREFUL WITH THIS! 
    * There is no recovery once this is called.
    */
   public void clearAnnotations()
   {
      annotations = null;
   }
   
   /**
    * Get an iterator on the keys for all annotations (if any) stored in this
    * AST.
    *
    * @return  An iterator on all annotation keys. May be empty, but will not
    *          be <code>null</code>.
    */
   @Override
   public Iterator<String> annotationKeys()
   {
      if (annotations == null)
      {
         return new EmptyIterator<>();
      }
      
      return annotations.keySet().iterator();
   }
   
   /**
    * Assigns parent nodes throughout the AST hierarchy of which this AST is
    * the root. This method should be called directly after creation of an
    * AST hierarchy, but before that hierarchy is post-processed. It
    * recursively walks downward through the tree, assigning itself as the
    * parent to each of its children.
    * <p>
    * If the current node's ID is <code>null</code>, then this method will
    * use the {@link AstManager#getNextNodeId} to obtain a valid ID that 
    * will uniquely identify this node. This can only be done if the passed
    * in file ID is valid and not <code>null</code>.
    * <p>
    * No shadow nodes will be created using this version of the method. All
    * hidden tokens will be dropped.
    *
    * @param   parent
    *          Parent AST node to be assigned to the current node.
    * @param   fileId
    *          The file ID of the source file.
    */
   public void fixups(Aast parent, Long fileId)
   throws AstException
   {
      fixups(parent, fileId, false);
   }
   
   /**
    * Assigns parent nodes throughout the AST hierarchy of which this AST is
    * the root. This method should be called directly after creation of an
    * AST hierarchy, but before that hierarchy is post-processed. It
    * recursively walks downward through the tree, assigning itself as the
    * parent to each of its children.
    * <p>
    * If the current node's ID is <code>null</code>, then this method will
    * use the {@link AstManager#getNextNodeId} to obtain a valid ID that 
    * will uniquely identify this node. This can only be done if the passed
    * in file ID is valid and not <code>null</code>.
    *
    * @param   parent
    *          Parent AST node to be assigned to the current node.
    * @param   fileId
    *          The file ID of the source file.
    * @param   shadow
    *          <code>true</code> to create shadow nodes from hidden tokens.
    *          <code>false</code> to drop hidden tokens without creating any
    *          shadow nodes.
    */
   public void fixups(Aast parent, Long fileId, boolean shadow)
   throws AstException
   {
      this.parent = parent;
      
      if (getId() == null && fileId != null)
      {
         setId(AstManager.get().getNextNodeId(fileId));
      }
      
      if (fileId != null && shadow)
      {
         // deferred processing of any hidden tokens that were cached off
         // during AST creation (must be after ID and parent are assigned)
         createShadowNodes(fileId);
      }
      
      AnnotatedAst child = (AnnotatedAst) getFirstChild();
      while (child != null)
      {
         child.fixups(this, fileId, shadow);
         child = (AnnotatedAst) child.getNextSibling();
      }
   }
   
   /**
    * Create a string representation of this AST. This implementation simply
    * uses the token name associated with this AST's token type.
    *
    * @return  Token name for this AST's token type.
    */
   public String toString()
   {
      return getSymbolicTokenType();
   }
   
   /**
    * Create a verbose string representation of this AST.  This includes the
    * the token name associated with this AST's token type, its text, its
    * line and column number, and its ID if non-null.
    *
    * @return  Verbose description of this node.
    */
   public String toStringVerbose()
   {
      StringBuilder sb = new StringBuilder(getText());
      
      // append the token type as a name
      sb.append(" [").append(getSymbolicTokenType()).append("] ");
      
      // append token position
      sb.append("@").append(getLine()).append(':');
      sb.append(getColumn());
      
      // append ID if any has been assigned
      if (id != null)
      {
         sb.append(" (").append(id).append(")");
      }
      
      return sb.toString();
   }
   
   /**
    * Create a human readable text representation of the tree rooted at this
    * AST. This implementation walks the subtree and dumps the text, token
    * type, line and column of each node, with indents to show the nesting 
    * and each node on its own line.
    *
    * @return  The tree's text representation.
    */
   public String dumpTree()
   {
      ByteArrayOutputStream bs = new ByteArrayOutputStream(); 
      PrintStream           ps = new PrintStream(bs, true);
      DumpTree              v  = new DumpTree(ps);
      
      v.visit(this);
      return bs.toString();
   }
   
   /**
    * Create a human readable text representation of the tree rooted at this AST. This
    * implementation walks the subtree and dumps the text, token type, line and column of each
    * node, with indents to show the nesting and each node on its own line.
    * 
    * @param   annotations
    *          Verbose mode. If {@code true} dump the annotations along with node name and type.
    *
    * @return  The tree's text representation.
    */
   public String dumpTree(boolean annotations)
   {
      ByteArrayOutputStream bs = new ByteArrayOutputStream(); 
      PrintStream           ps = new PrintStream(bs, true);
      DumpTree              v  = new DumpTree(ps);
      
      v.setDumpAnnotations(annotations);
      v.visit(this);
      return bs.toString();
   }
   
   /**
    * Access a shadow node that exists to the left of the current node.
    *
    * @return   The shadow node that is left of the current node or 
    *           <code>null</code> if no such node exists.
    */
   public Aast getLeftShadowNode()
   {
      // use getLocal() since the tree-local reference may only be stored in
      // the root node
      TreeLocal  local = getLocal();
      ShadowNode left  = (local == null) ? null : local.getLeftShadowNode(id);  
      return left == null ? null : left.getNode();
   }
   
   /**
    * Access a shadow node that exists to the right of the current node.
    *
    * @return   The shadow node that is right of the current node or 
    *           <code>null</code> if no such node exists.
    */
   public Aast getRightShadowNode()
   {
      // use getLocal() since the tree-local reference may only be stored in
      // the root node
      TreeLocal  local = getLocal();
      ShadowNode right = (local == null) ? null : local.getRightShadowNode(id);  
      return right == null ? null : right.getNode();
   }
   
   /**
    * Access the list of all shadow nodes.
    *
    * @return   All shadow nodes or <code>null</code> if none exist.
    */
   public Iterator<ShadowNode> getShadowNodeList()
   {
      // use getLocal() since the tree-local reference may only be stored in
      // the root node
      TreeLocal local = getLocal();
      return local == null ? null : local.getShadowNodeList();
   }
   
   /**
    * Access the tree-local data directly.  WARNING: this data is mutable so
    * make sure you really want to change it. If the tree is subsequently
    * persisted, any changes will be saved.
    *
    * @return   The tree-local data or <code>null</code> if there is no
    *           tree-local data.
    */
   public TreeLocal getLocal()
   {
      TreeLocal local = this.local;
      
      // look up the tree to find the local data (it may only be set in the
      // root node)
      if (local == null && !isRoot())
      {
         Aast root = getAncestor(-1);
         local = root.getLocal();
      }
      
      return local;
   }
   
   /**
    * Set the tree-local data directly.  WARNING: this data is mutable so
    * make sure you really want to change it. If the tree is subsequently
    * persisted, any changes will be saved.
    *
    * @param    local
    *           The tree-local data or <code>null</code> if there is no
    *           tree-local data.
    */
   public void setLocal(TreeLocal local)
   {
      // only store the local data reference in the root node, getLocal()
      // knows to look there and this limits the number of places where we
      // must copy the reference, null the reference...
      if (isRoot())
      {
         this.local = local;
      }
      else
      {
         Aast root = getAncestor(-1);
         root.setLocal(local);
      }
   }
   
   /**
    * Create a new AST node of the same sub-class as the current node and
    * initialize it using the given token.
    *
    * @param    token
    *           The data with which to populate the node.
    * @param    parent
    *           The parent for the newly created node.
    *
    * @return   The newly created node or <code>null</code> if there is an
    *           error.
    */
   private AnnotatedAst createNode(Token token, Aast parent)
   {
      AnnotatedAst node = null;
      
      try
      {
         node = this.getClass().newInstance();
      }
      catch (Exception exc)
      {
         // nothing to do but return
         return null;
      }
      
      node.initialize(token);
      node.setParent(parent);
      
      return node;
   }
   
   /**
    * Get chain of parent nodes starting from root to this node.
    * 
    * @return  array of parent nodes.
    */
   private Aast[] getParentChain(Aast node)
   {
      List<Aast> res = new ArrayList<>();
      
      while (node != null)
      {
         res.add(0, node);
         node = node.getParent();
      }
      
      return res.toArray(new Aast[res.size()]);
   }
   
   /**
    * Check the annotations for the special <code>hidden-before</code> and
    * <code>hidden-after</code> hidden token objects and create shadow nodes
    * from these tokens if they exist. The entire chain of hidden tokens will
    * be walked and created.  Each shadow node will have a valid ID and will
    * have the left and right IDs set as appropriate.  Then the cached
    * annotations will be removed so that there is never an attempt to persist
    * those to file.
    * <p>
    * Both the parent and the ID must be assigned before this method is called.
    *
    * @param   fileId
    *          The file ID of the source file.
    */
   private void createShadowNodes(Long fileId)
   {
      AstManager mgr  = null;
      Aast       root = getAncestor(-1);
      
      // use getLocal() since the tree-local reference may only be stored in
      // the root node
      TreeLocal local = getLocal();
      
      // remember if the local data was null on entry
      boolean assign = (local == null);
      
      CommonHiddenStreamToken b4 = 
         (CommonHiddenStreamToken) removeAnnotation("hidden-before");
      
      if (b4 != null)
      {
         if (local == null)
            local = new TreeLocal();
         
         int                     idx    = local.shadowListSize();
         ShadowNode              last   = local.getShadowNode(idx - 1);
         CommonHiddenStreamToken source = (last != null) ? last.getSource() : null;
         
         if (getPrevSibling() == null)
         {
            Long       rtId   = getId();
            ShadowNode rtNode = null;
            
            mgr = AstManager.get();
            
            // first child processes all hidden tokens on our left
            while (b4 != null)
            {
               ShadowNode prev = new ShadowNode();
               
               // create the new AST and get a new id for it
               prev.setNode(createNode(b4, root));
               prev.setSource(b4);
               Long ltId = mgr.getNextNodeId(fileId);
               prev.getNode().setId(ltId);
               
               // setup its right-side linkage 
               prev.setRight(rtId);
               rtId = ltId;
               
               // setup the left-side linkage of the previous shadow node if
               // there was one
               if (rtNode != null)
                  rtNode.setLeft(ltId);
               
               // add the shadow node to our list, but insert at the same
               // location each time; first time through it is at the end,
               // second time through it is next to last, third time through it
               // is 2nd to last...; this is important to retain the proper
               // order of the nodes
               local.addShadowNode(idx, prev);
               
               // remember the current shadow node as the previous one so that
               // its left-side linkage can be fixed up next pass
               rtNode = prev;
               
               b4 = b4.getHiddenBefore();
               
               if (b4 != null)
               {
                  ShadowNode old = local.getShadowNode(b4);
                  
                  if (old != null)
                  {
                     // this is a case where we have must connect back to shadow
                     // nodes that were already created in another part of the
                     // tree (otherwise we would create duplicate AST nodes)
                     old.setRight(rtId);
                     
                     // no reason to continue
                     break;
                  }
               }
            }
         }
         else
         {
            // just fixup the nearest connection, which must be the last shadow
            // node in our list since this processing is driven during a normal
            // depth-first, left to right tree walk
            last.setRight(id);
         }
      }
      
      CommonHiddenStreamToken after = 
         (CommonHiddenStreamToken) removeAnnotation("hidden-after");
      
      if (after != null)
      {
         Long       ltId   = getId();
         ShadowNode ltNode = null;
         
         if (mgr == null)
         {
            mgr = AstManager.get();
         }
         
         // always process all hidden nodes to the right
         while (after != null)
         {
            if (local == null)
            {
               local = new TreeLocal();
            }
            
            ShadowNode next = new ShadowNode();
                  
            // create the new AST and get a new id for it
            next.setNode(createNode(after, root));
            next.setSource(after);
            Long rtId = mgr.getNextNodeId(fileId);
            next.getNode().setId(rtId);
            
            // setup its left-side linkage 
            next.setLeft(ltId);
            ltId = rtId;
            
            // setup the right-side linkage of the previous shadow node if
            // there was one
            if (ltNode != null)
            {
               ltNode.setRight(rtId);
            }
            
            // add the shadow node to our list
            local.addShadowNode(next);
            
            // remember the current shadow node as the previous one so that
            // its right-side linkage can be fixed up next pass
            ltNode = next;
            
            after = after.getHiddenAfter();
         }
      }
      
      // save off our tree local data if we were the first to create it
      if (local != null && assign)
      {
         setLocal(local);
      }
   }
   
   private boolean searchBack(TreeLocal local, CommonHiddenStreamToken tok)
   {
      int size = local.shadowListSize();
      
      for (int idx = size - 1; idx >= 0; idx--)
      {
         ShadowNode last = local.getShadowNode(idx);
         
         CommonHiddenStreamToken source = (last != null) ? last.getSource() : null;
         if (source == tok || tokensMatch(source, tok))
         {
            System.out.println("match idx " + idx + " when size is " + size);
            System.out.println("source " + source + " tok " + tok); 
         }
      }
      
      return true;
   }
   
   /**
    * 
    */
   private boolean tokensMatch(CommonToken tok1, CommonToken tok2)
   {
      if (tok1 == null || tok2 == null)
         return false;
      
      String t1 = tok1.getText();
      String t2 = tok2.getText();
      
      if (t1 == null)
         t1 = "";
      else
         if (t1.contains("even numbers greater than 2 can't be prime!"))
            System.out.println("1 " + t1 + "   " + t2);
      
      if (t2 == null)
         t2 = "";
      else
         if (t2.contains("even numbers greater than 2 can't be prime!"))
            System.out.println("2 " + t1 + "   " + t2);
      
      return t1.equals(t2)                        &&
             tok1.getLine() == tok2.getLine()     &&
             tok1.getColumn() == tok2.getColumn();
   }
   
   /**
    * Return a string representing the path of this AST from an ancestor (possibly the root) AST,
    * in terms of the token type of each element from the ancestor to this AST. The token types
    * are separated by a forward slash and a trailing slash ends the path. For example:
    * <pre>
    *    BLOCK/STATEMENT/KW_IF/
    * </pre>
    * represents the following path:
    * <pre>
    *    ROOT [type=BLOCK]
    *     |
    *    NODE A [type=STATEMENT]
    *     |
    *    NODE B [type=KW_IF]
    * </pre>
    * The string uniquely identifies the path from this node back to the ancestor node, but is not
    * necessarily unique when traversed from the ancestor to this node. How far up the tree we
    * travel is determined by the <code>max</code> parameter.
    * 
    * @param   max
    *          The maximum number of characters to be included in the returned path, or a
    *          non-positive integer to indicate that all nodes up to and including the root
    *          should be included in the path.
    *
    * @return  A string which, when read from right to left, indicates the token type of each node
    *          from this AST to the ancestor AST, or if <code>max &gt; 0</code>, as much of that
    *          path that can fit within <code>max</code> characters.
    */
   private String getPath(int max)
   {
      int size = 0;
      List<String> list = new ArrayList<>();
      Aast next = this;
      do
      {
         String s = next.getSymbolicTokenType();
         list.add(s);
         size += (s.length() + 1);
      } while ((max <= 0 || size < max) && (next = next.getParent()) != null);
      
      String sep = "/";
      StringBuilder buf = new StringBuilder(size);
      
      ListIterator<String> iter = list.listIterator(list.size());
      while (iter.hasPrevious())
      {
         String s = iter.previous();
         buf.append(s);
         buf.append(sep);
      }
      
      return buf.toString();
   }
   
   /**
    * Returns the first child (hidden or not) if the <code>honorHidden</code>
    * flag is <code>false</code>, or the first non-hidden child if the
    * <code>honorHidden</code> flag is <code>true</code>.
    *
    * @param   honorHidden
    *          If <code>true</code>, the <code>hidden</code> attribute of
    *          the node will be honored.  This means that the node will not
    *          be returned and instead the first non-hidden sibling will be
    *          returned.
    *
    * @return  The first non-hidden child (in <code>honorHidden</code> mode),
    *          or otherwise the first child. Returns <code>null</code> if no
    *          non-hidden children exist (in <code>honorHidden</code> mode)
    *          or whenever no children of any kind exist.
    */
   private Aast getFirstChild(boolean honorHidden)
   {
      Aast child = (Aast) getFirstChild();
      
      if (honorHidden && child != null && child.isHidden())
      {
         child = ((AnnotatedAst) child).getNextSibling(honorHidden);
      }
      
      return child;
   }
   
   /**
    * Returns the next sibling (hidden or not) if the <code>honorHidden</code>
    * flag is <code>false</code>, or the next non-hidden sibling if the
    * <code>honorHidden</code> flag is <code>true</code>.
    *
    * @param   honorHidden
    *          If <code>true</code>, the <code>hidden</code> attribute of
    *          the node will be honored.  This means that the node will not
    *          be returned and instead the next non-hidden sibling will be
    *          returned.
    *
    * @return  The next non-hidden sibling (in <code>honorHidden</code> mode),
    *          or otherwise the next child. Returns <code>null</code> if no
    *          non-hidden siblings exist (in <code>honorHidden</code> mode)
    *          or whenever no siblings of any kind exist.
    */
   private Aast getNextSibling(boolean honorHidden)
   {
      Aast child = (Aast) getNextSibling();
      
      while (honorHidden && child != null && child.isHidden())
      {
         child = (Aast) child.getNextSibling();
      }
      
      return child;
   }

   /**
    * Worker to recursively search down the descendant tree to match the
    * passed-in list of token names.
    *
    * @param    list
    *           The list of token names in order from ancestor (leftmost)
    *           to descendant (rightmost).
    * @param    index
    *           The current segment of the path which must be tested.
    *
    * @return   If the path matches completely <code>true</code>.
    */
   private boolean searchDown(String[] list, int index)
   {
      // test to see if this is the last segment of the list to check
      boolean end = ((index + 1) == list.length);
      
      // convert the segment name into a token type
      int type = lookupTokenType(list[index]);
      
      // confirm the segment is valid
      if (type == -1)
      {
         // invalid type (GIGO)
         return false;
      }
      
      // get first match to this segment
      AnnotatedAst child = (AnnotatedAst) getImmediateChild(type, null);
      
      while (child != null)
      {
         // if this is the end of the down path return true OTHERWISE 
         // recursively call down the descendants to check
         if (end || child.searchDown(list, index + 1)) 
         {
            // this is the last segment and it matches OR the descendants
            // all match
            return true;
         }
            
         // get next match to this segment
         child = (AnnotatedAst) getImmediateChild(type, child);
      }
      
      return false;
   }
   
   /**
    * Worker to compare the given source path (from root to a given node) to
    * the expected path to the target and return the comparison results.
    * Please note that a match exists if the source string ends with the
    * target string on a case-insensitive basis.  Some simple normalization
    * of path strings is done before comparison to ensure that the format is
    * consistent.
    *
    * @param    source
    *           The actual path upward.
    * @param    target
    *           The comparison path.
    *
    * @return   <code>true</code> if the source path ends with the target 
    *           path.
    */
   private boolean searchUp(String source, String target)
   {
      // paths should always end with a '/' but we will make up for a missing one here
      if (!target.endsWith("/"))
      {
         target = target + "/";
      }
      
      // calculate the beginning index of the relative match
      int index = source.length() - target.length();
      
      // we can't have a match if the requested path is longer than the original
      if (index < 0)
      {
         return false;
      }
      
      // subset the path to match against
      String end = source.substring(index);
      
      // test and return the result 
      return end.equalsIgnoreCase(target);
   }
   
   /**
    * Kickoff the fixup process on a specific child to set IDs and ensure
    * that all parent linkages are correct.
    *
    * @param    child
    *           The child node to fixup.
    */
   private void driveFixups(Aast child)
   throws AstException
   {
      Long id     = getId();
      Long treeId = null;
      
      if (id != null)
      {
         AstManager mgr = AstManager.get();
         treeId = mgr.getTreeId(getId());
      }
      
      // fixup parent links and the ids
      ((AnnotatedAst)child).fixups(this, treeId);
   }
   
   /**
    * Worker method to handle the core implementation of the 'deep copying'
    * in the copy rather than in the original AST.  This method will be
    * called on the copy AST by the original AST.  This is the place where
    * recursion occurs, allowing all children to be duplicated.
    *
    * @param   original
    *          AST to be copied.
    * @param   idMap
    *          Map of IDs to ASTs. May be <code>null</code>.
    */
   private void duplicateImpl(AnnotatedAst original, Map<Long, Aast> idMap)
   {
      initialize(original);
      
      // Map this node to its ID if a map is provided and an id was copied.
      if (idMap != null && id != null)
      {
         idMap.put(id, this);
      }
      
      // Copy children recursively, fixing up parent references as we go.
      AnnotatedAst origChild = (AnnotatedAst) original.getFirstChild();
      while (origChild != null)
      {
         AnnotatedAst copyChild = (AnnotatedAst) origChild.duplicate(idMap);
         copyChild.parent = this;
         addChild(copyChild);
         origChild = (AnnotatedAst) origChild.getNextSibling();
      }
   }   
   
   /**
    * Stores an annotation object under a unique key, creating the map of
    * annotations first, if necessary.
    *
    * @param   key
    *          Key under which to store the annotation.
    * @param   annotation
    *          Arbitrary annotation value.
    */
   private void putAnnotationImpl(String key, Object annotation)
   {
      if (annotations == null)
      {
         annotations = new LinkedHashMap<>(4);
      }
      
      if (poolStrings && annotation instanceof String)
      {
         annotation = ((String) annotation).intern();
      }
      
      annotations.put(key.toLowerCase().intern(), annotation);
   }
   
   /**
    * Stores an annotation object in an array annotation under a unique key,
    * creating both the array itself and the map of annotations first, if
    * necessary.
    *
    * @param   key
    *          Key under which to store the annotation.
    * @param   annotation
    *          Arbitrary annotation value.
    * @param   index
    *          The index of -1 indicates an addition to the end of the array.
    *          Otherwise this is the explicit zero-based index at which to 
    *          add the annotation.  Any element at that location will be
    *          replaced.
    *
    * @return  <code>true</code> if the element was added.
    */
   private boolean putAnnotationImpl(String key, Object annotation, int index)
   {
      if (poolStrings && annotation instanceof String)
      {
         annotation = ((String) annotation).intern();
      }
      
      Object obj = getAnnotation(key);
      if (annotations == null)
      {
         annotations = new LinkedHashMap<>(4);
      }
      
      List<Object> array;
      if (obj == null)
      {
         // create our array if it doesn't exist already
         array = new ArrayList<>(1);
         annotations.put(key.toLowerCase().intern(), array);
      }
      else
      { 
         if (!(obj instanceof List<?>))
         {
            // instead of throwing an exception, we just silently stop
            // this was done to ease the use of these interfaces (but at the
            // expense of easy problem determination)
            return false;
         }
                  
         array = (List<Object>) obj;
      }
      
      if (index == -1)
      {
         // set the last element in the array
         array.add(annotation);
      }
      else
      {
         try
         {
            if (index >= 0 && index < array.size())
            {
               // there is an element already at that position in the array so remove it
               array.remove(index);
            }
            
            array.add(index, annotation); 
         }
         catch (IndexOutOfBoundsException iob)
         {
            // bad index
            return false;
         }
      }
      
      return true;
   }
   
   /**
    * Extract line number, column number and hidden previous/next token data
    * from the input token and store this information as member data. The
    * hidden token data will only be cached here if the input token class
    * is a <code>CommonHiddenStreamToken</code>.  Otherwise, that data will
    * be ignored.
    * <p>
    * The hidden previous/next token data will be temporarily stored using
    * the annotations <code>hidden-before</code> and <code>hidden-after</code>
    * before being moved into true shadow nodes by the brainwashing process.
    * This is a hack but it is the simplest and cleanest approach available
    * until better hooks exist for processing IDs and parent linkages during
    * node creation.  Since the shadow nodes cannot be fully processed until
    * the node IDs are assigned, the hidden token data must be cached with the
    * node until that time.  Then the temporary annotations will be removed
    * and the shadow nodes will be created and stored in the tree-local data
    * store.  These temporary annotations will never be persisted to disk.
    *
    * @param   token
    *          Token from which information is to be extracted.
    */
   private void extractTokenData(Token token)
   {
      this.line = token.getLine();
      this.column = token.getColumn();
      
      // temporarily cache hidden tokens until brainwash can process these
      if (token instanceof CommonHiddenStreamToken)
      {
         CommonHiddenStreamToken chtok = (CommonHiddenStreamToken) token;
         
         Token b4 = chtok.getHiddenBefore();
         if (b4 != null)
         {
            this.putAnnotationImpl("hidden-before", b4);
         }
         
         Token after = chtok.getHiddenAfter();
         if (after != null)
         {
            this.putAnnotationImpl("hidden-after", after);
         }
      }
   }
   
   /**
    * Compare subtree starting from <code>this</code> node as a root and
    * compare it with given subtree. Subtrees considered identical if both
    * contain identical nodes organized in the same order. Two nodes are
    * considered identical if their type and text are identical. The order of
    * subtrees is defined by iterator provided by iterator() method.
    * <p>
    * The comparison of token text is case sensitive.  For a case insensitive
    * comparison, use the {@link #compareWith(Aast,boolean)} method variant.
    * 
    * @param   other
    *          The subtree to compare with.
    * 
    * @return  <code>true</code> if subtrees are identical and
    *          <code>false</code> otherwise.
    */
   public boolean compareWith(Aast other)
   {
      return compareWith(other, false);
   }
   
   /**
    * Compare subtree starting from <code>this</code> node as a root and
    * compare it with given subtree. Subtrees considered identical if both
    * contain identical nodes organized in the same order. Two nodes are
    * considered identical if their type and text are identical. The order of
    * subtrees is defined by iterator provided by iterator() method.
    * 
    * @param   other
    *          The subtree to compare with.
    * @param   ignoreCase
    *          <code>true</code> to ignore case in the token text comparison;
    *          <code>false</code> for a case sensitive comparison.
    * 
    * @return  <code>true</code> if subtrees are identical and
    *          <code>false</code> otherwise.
    */
   public boolean compareWith(Aast other, boolean ignoreCase)
   {
      if (other == null)
      {
         return false;  // any non-empty subtree is not equal to empty subtree
      }
      
      Iterator<Aast> iterSrc = iterator();
      Iterator<Aast> iterDst = other.iterator();
      
      while (iterSrc.hasNext() && iterDst.hasNext())
      {
         Aast astSrc = iterSrc.next();
         Aast astDst = iterDst.next();
         String textSrc = astSrc.getText();
         String textDst = astDst.getText();
         if (ignoreCase)
         {
            textSrc = textSrc.toLowerCase();
            textDst = textDst.toLowerCase();
         }
         
         if (astSrc.getType() != astDst.getType() || !textSrc.equals(textDst))
         {
            return false;
         }
      }
      
      // if anything is left non-compared, trees are not equal
      return !(iterSrc.hasNext() || iterDst.hasNext());
   }
}
